/* Copyright 2007 Deutsches Zentrum fuer Luft- und Raumfahrt e.V. (DLR)
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#include <httpd.h>
#include <http_protocol.h>
#include <http_log.h>
#include <http_core.h>      /* for ap_construct_url */
#include <mod_dav.h>

#include <apr.h>
#include <apr_strings.h>
#include <apr_hash.h>
#include <apr_uuid.h>
#include <apr_file_io.h>
#include <apr_lib.h>

#include <stdio.h>

#include "dav_repos.h"
#include "lock.h"
#include "dbms.h"
#include "util.h"

/* EXPERIMENTAL START */
#include <apr_dbd.h>
#include <mod_dbd.h>
#include <apr_base64.h>

#define BINFLAG APR_CREATE | APR_WRITE | APR_TRUNCATE | APR_BINARY | APR_BUFFERED
#define TEXTFLAG APR_CREATE | APR_WRITE | APR_TRUNCATE | APR_BUFFERED
/* EXPERIMENTAL END */

#define DASL_MYSQL_DEBUG

/**
 * Show error message of SQL operations
 * @param db_error_message_str An application level error message string
 * @param db_errno The DB error number
 * @param dberrmessage The DB error message from the DB
 *//*
void static db_error_message ( apr_pool_t *pool, const dav_repos_dbms *db,
                               char *db_error_message_str )
{
    ap_log_error ( APLOG_MARK,
                   APLOG_ERR, 0, NULL,
                   "Error: %s\nError Number: %d\nErrorMessage:%s\n",
                   db_error_message_str, dbms_errno ( db ),
                   dbms_error ( pool, db ) );
}*/

/* EXPERIMENTAL START */

/**
 * Show error message of SQL operations
 * @param db_error_message_str An application level error message string
 * @param db_errno The DB error number
 * @param dbd The database connection/handle
 */
void db_error_message_new ( ap_dbd_t *dbd, const int db_errno,
                               const char *db_error_message_str, const char* func )
{
    ap_log_error ( APLOG_MARK,
                   APLOG_ERR, 0, NULL,
                   "Error: %s\nError Number: %d\nErrorMessage: %s - %s\n",
                   apr_dbd_error(dbd->driver, dbd->handle, db_errno), 
                   db_errno, 
                   db_error_message_str,
                   func);
}

/* EXPERIMENTAL END */

/**
 * Given a particular resource (with it's serialno, type, etc) return a
 * path into the filesystem for storage. Change the algorithm to change
 * the layout of how data is stored in the fs. *path is set to NULL if
 * there is no file_dir defined (as the root dir.)
 */
static dav_error *generate_path ( char **path, apr_pool_t *pool,
                                  const char *file_dir, const long serialno, const int version )
{
    char *dirpath;

    TRACE();

    DBG2 ( "Serial number: %ld version: %d", serialno, version );
    dirpath = apr_psprintf ( pool, "%s/%ld",
                             file_dir, serialno );
    ap_no2slash ( dirpath );

    // should I check to see if dirpath exists already or can I assume
    // apr_dir_make_recursive is smart 'nuf?
    if ( apr_dir_make_recursive ( dirpath, APR_OS_DEFAULT, pool ) != APR_SUCCESS )
        return dav_new_error ( pool, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "Unable to create directory for storage." );

    *path = apr_psprintf ( pool, "%s/%ld-%d", dirpath, serialno, version );
    return NULL;
}

/**
 * Connects to and opens the database
 * @param d DB connection struct containing the user, password, and DB name
 * @return 0 indicating success
 *//*
int dbms_opendb ( dav_repos_db * d )
{
    TRACE();

    dbms_create ( & ( d->db ), d->db_id, d->db_pass, d->db_name, d->db_host, 0 );
    return 0;
}*/

/**
 * Disconnect from the database server
 * @param d DB connection struct containing the user, password, and DB name
 * @return 0 indicating success
 *//*
int dbms_closedb ( dav_repos_db * d )
{
    TRACE();

    dbms_destroy ( & ( d->db ) );
    return 0;
}*/

/**
 * r->uri for search key
 * @param d DB connection struct containing the user, password, and DB name
 * @param r Out variable to hold the result
 * @return 1 on success, 0 for no result, -1 on fail
 */
int dbms_get_property ( dav_repos_db * d, dav_repos_resource * r )
{
    int is_vhr=0, is_vr=0;
    char *uri;
    apr_pool_t *pool = r->p;

    /* EXPERIMENTAL START */
    apr_status_t rv = 0;
    int first = 1;
    /* mod_dbd vars */
	ap_dbd_t *dbd = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	int ret_val = 0;
    /* EXPERIMENTAL END */
    
    TRACE();

    /* Let's check the URI is VCR or VHR */
    uri =  dav_repos_parse_vcr_uri ( pool, r->uri, & ( r->version ), &is_vhr );

    if ( uri == NULL ) /* Nor VR or VHR */
        uri = r->uri;
    else if ( is_vhr==0 )
        is_vr = 1;

    /* If it is VCR, you want to query version_resource using
       uri and vnum
    */

    /* EXPERIMENTAL START */
    dbd = dbms_get_db(d);
	/* EXPERIMENTAL END */
    
    if ( is_vr == 0 )
    { /* It is not VCR */
  	
    	/* EXPERIMENTAL START */
    	if(dbd) {
    		ret_val = apr_dbd_prepare(dbd->driver, pool, dbd->handle,"SELECT serialno,"
			                           "creationdate,"
			                           "displayname,"
			                           "getcontentlanguage,"
			                           "getcontentlength,"
			                           "getcontenttype,"
			                           "getetag,"
			                           "getlastmodified,"
			                           "resourcetype,"
			                           "source,"
			                           "depth,"
			                           "istext, "
			                           "isexternal, "
			                           "isversioned, "
			                           "checkin, "
			                           "checkout "
			                           "FROM dasl_resource WHERE URI = %s",
			                           NULL,
			                           &pselect);
    		
    		if(ret_val) {
    			db_error_message_new(dbd, ret_val, "Error mod_dbd", __func__);
				return -1;
			}

    		ret_val = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1, uri, NULL);
    	}
    	/* EXPERIMENTAL END */
    	
    }
    else
    { /* It is VCR */
    	
    	/* EXPERIMENTAL START */
		if(dbd) {
			
			ret_val = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT vr.serialno,"
					                    "vr.creationdate,"
					                    "vr.displayname,"
					                    "vr.getcontentlanguage,"
					                    "vr.getcontentlength,"
					                    "vr.getcontenttype,"
					                    "vr.getetag,"
					                    "vr.getlastmodified,"
					                    "vr.resourcetype,"
					                    "vr.source,"
					                    "vr.depth,"
					                    "vr.istext, "
					                    "vr.isexternal "
					                    "FROM dasl_resource, version_resource vr "
					                    "WHERE dasl_resource.URI= %s AND dasl_resource.serialno=vr.serialno AND "
					                    "vr.version= %d ",
					                    NULL,
					                    &pselect);
			
	    	if(ret_val) {
	    		db_error_message_new(dbd, ret_val, "Error mod_dbd", __func__);
				return -1;
			}
	
			rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1, uri, r->version, NULL);

		}
		/* EXPERIMENTAL END */
    }

    /* EXPERIMENTAL START */
	
    
    
    if(!ret_val) {
    	for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
    		
    		if(rv != 0) {
				db_error_message_new(dbd, rv, "Error mod_dbd",  __func__);
    		}
    		
    		if(first) {
    			
    			r->serialno = atoll( apr_dbd_get_entry(dbd->driver, row, 0 ));
    		    /*r->filename = dbms_get_string(q, 2); For Catacomb-v*/
    		    r->creationdate = atoll( apr_dbd_get_entry(dbd->driver, row, 1 ));
    		    r->displayname = apr_dbd_get_entry(dbd->driver, row, 2 );
    		    r->getcontentlanguage = apr_dbd_get_entry(dbd->driver, row, 3 );
    		    r->getcontentlength = atoll( apr_dbd_get_entry(dbd->driver, row, 4 ));
    		    r->getcontenttype = apr_dbd_get_entry(dbd->driver, row, 5 );
    		    r->getetag = apr_dbd_get_entry(dbd->driver, row, 6 );
    		    r->getlastmodified = atoll( apr_dbd_get_entry(dbd->driver, row, 7 ));
    		    r->resourcetype = atoll( apr_dbd_get_entry(dbd->driver, row, 8 ));
    		    r->source = atoll( apr_dbd_get_entry(dbd->driver, row, 9 ));
    		    r->depth = atoll( apr_dbd_get_entry(dbd->driver, row, 10 ));
    		    r->istext = atoll( apr_dbd_get_entry(dbd->driver, row, 11 ));
    		    r->isexternal = atoll( apr_dbd_get_entry(dbd->driver, row, 12 ));

    		    if ( is_vr == 0 )
    		    {
    		        r->isversioned = atoll( apr_dbd_get_entry(dbd->driver, row, 13 ));
    		        r->checkin = atoll( apr_dbd_get_entry(dbd->driver, row, 14 ));
    		        r->checkout = atoll( apr_dbd_get_entry(dbd->driver, row, 15 ));
    		    }
    		    
    		    first = 0;
    		}
    	}
	}
    
    if (ret_val != 0) {
		return -1;
	}
    
    if(first) {
		return 0;
	}
    
	/* EXPERIMENTAL END */ 

    r->next = NULL;

    return 1;
}               /*End of dbms_get_property */

/**
 * Inserts a resource into the dasl_resource table
 * @param d DB connection struct containing the user, password, and DB name
 * @param r The resource to insert
 * @param lserialno The serial number assigned to the resource when inserted
 * @return 0 if success, -1 if fail
 */
int db_insert_resource ( dav_repos_db * d,
                         const dav_repos_resource * r, long *lserialno )
{
    apr_pool_t *pool = r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	int rv = 0;
	int nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();
    
    /* EXPERIMENTAL START */
    dbd = dbms_get_db(d);

	if(dbd) {
	    rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "INSERT INTO dasl_resource \
	                       (uri,\
	                       creationdate,\
	                       displayname,\
	                       getcontentlanguage,\
	                       getcontentlength,\
	                       getcontenttype,\
	                       getetag,\
	                       getlastmodified,\
	                       resourcetype,\
	                       source,\
	                       depth, \
	                       istext, \
	                       isexternal ) \
	                       VALUES(%s, %d, %s, %s, %d, %s,\
	                       %s, %d, %d, %d, %d, %d, %d)", 
	                       NULL, &pquery);
		
	    if(rv != 0) {
	    	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
			return rv;
		}
		
		rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery,
				r->uri,
				dav_repos_lltoa(pool ,r->creationdate),
				apr_filepath_name_get ( r->uri ),
				dav_repos_dbms_string(r->getcontentlanguage),
				apr_ltoa(pool, r->getcontentlength),
				dav_repos_dbms_string(r->getcontenttype),
				dav_repos_dbms_string(r->getetag),
				dav_repos_lltoa(pool ,r->getlastmodified),
				apr_itoa(pool, r->resourcetype),
				apr_itoa(pool, r->source),
				apr_itoa(pool, r->depth),
				apr_itoa(pool, r->istext),
				apr_itoa(pool, r->isexternal));
		if(rv != 0) {
			db_error_message_new(dbd, rv, "Insert resource failed", __func__);
			return rv;
		}
		
		*lserialno = dbms_insert_id(d, "dasl_resource", "serialno");
		DBG1 ( "SERIAL=%p", &lserialno );
	}
	/* EXPERIMENTAL END */
    
    return 0;
}               /*End of db_insert_resource */

/**
 * Set a property on a resource
 * @param d DB connection struct containing the user, password, and DB name
 * @param r Identifies the resource to set the property on r->uri.
 *          Contains the live property and value to set.
 * @return The serial number of the resource, or -1 on fail
 */

long dbms_set_property ( dav_repos_db * d, const dav_repos_resource * r )
{
    long lserialno = -1;
    apr_pool_t *pool = r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	int nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    /* EXPERIMENTAL START */
	dbd = dbms_get_db(d);
	
	if (dbd) {
		
		rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT serialno FROM dasl_resource"
                       " WHERE URI=%s", 
                       NULL, 
                       &pselect);
		
		if (rv != 0) {
			db_error_message_new(dbd, rv, "prepare statement failed", __func__);
			return -1;
		}
		
		rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1, r->uri);
		
		if (rv == 0) {
			for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
				rv != -1;
				rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
				
				lserialno = atol(apr_dbd_get_entry(dbd->driver, row, 0));
				
			}
		}
		else {
			db_error_message_new(dbd, rv, "apr_dbd execution failed", __func__);
			return -1;
		}
		
		if (lserialno == -1) {
			if ( db_insert_resource ( d, r, &lserialno ) != 0 )
			    return -1;
			else
			    return lserialno;
		}
		
		rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "UPDATE dasl_resource \
                       SET creationdate=%d,\
                       displayname=%s,\
                       getcontentlanguage=%s,\
                       getcontentlength=%d,\
                       getcontenttype=%s,\
                       getetag=%s,\
                       getlastmodified=%d,\
                       resourcetype=%d,\
                       source=%d,\
                       depth=%d,\
                       istext=%d, \
                       isexternal=%d \
                       WHERE uri=%s", NULL, &pquery);
		
		if (rv != 0) {
			db_error_message_new(dbd, rv, "prepare statement failed", __func__);
			return -1;
		}
		
		rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
				dav_repos_lltoa(pool, r->creationdate),
				apr_filepath_name_get( r->uri ),
				dav_repos_dbms_string(r->getcontentlanguage),
				apr_ltoa(pool, r->getcontentlength),
				dav_repos_dbms_string(r->getcontenttype),
				dav_repos_dbms_string(r->getetag),
				dav_repos_lltoa(pool, r->getlastmodified),
				apr_itoa(pool, r->resourcetype),
				apr_itoa(pool, r->source),
				apr_itoa(pool, r->depth),
				apr_itoa(pool, r->istext),
				((d->file_dir != NULL ) && ( d->max_file_size < r->getcontentlength ) ? "1" : "0"), 
				dav_repos_dbms_string(r->uri));
		
		if(rv != 0) {
			db_error_message_new(dbd, rv, "Update resource property failed", __func__);
			return -1;
		}
		
	}
	/* EXPERIMENTAL END */
    
    return lserialno;
}               /*End of dbms_set_property */

/**
 * Reads the text content to the buffer by serialno.
 * @param d DB connection struct containing the user, password, and DB name
 * @param bd_r Identifies the resource by bd_r->serialno
 * @return 0 on success, -1 on failure and 100 if not found
 */
int dbms_read_content ( dav_repos_db * d, dav_repos_resource * db_r,
                        const char *filename )
{
    apr_pool_t *pool = db_r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	apr_file_t *file = NULL;
	apr_int32_t flag = 0;
	int rv = 0;
	apr_size_t bytes = 0;
	apr_status_t code = 0;
	const char *data = NULL;
	const char *stmt = NULL;
	unsigned char *bindata;
	/* EXPERIMENTAL END */
    
    TRACE();
    
    dbd = dbms_get_db(d);
    
    if(dbd)
    {
	    if ( d->file_dir )
	    {}
	    /*Create the select command to get the textcontent by serialNO */
	    if ( db_r->type == DAV_RESOURCE_TYPE_VERSION )
	    { /* Version_resource */
	    	
	    	
	    	if ( db_r->istext == 1 )    /*It is a regular file */
	    	{
	    		flag = TEXTFLAG;
	    		stmt = "SELECT textcontent "
			    	   "FROM version_resource "
		               "WHERE serialno=%d AND version=%d";
	    		
	    	}
	        else if ( db_r->istext == 2 )   /*It is a binary file */
	        {
	        	flag = BINFLAG;
	        	stmt = "SELECT bincontent "
			    	   "FROM version_resource "
		               "WHERE serialno=%d AND version=%d";
	        		    		
	        }
	        else
	        { /* NONE */
	            DBG0 ( "istext is not 1 or 2." );
	            return 100;
	        }
	
	    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, stmt, 
			                     NULL, 
			                     &pselect);
				
			if(rv != 0) {   	
			   	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
				return -1;
			}
			
			/*Execute the select command */
			rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1,
					apr_ltoa(pool, db_r->serialno),
					apr_itoa(pool, db_r->version));
					
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
				return -1;
			}
	    }
	    else
	    { /* Normal resource*/
	        if ( db_r->istext == 1 )    /*It is a regular file */
	        {
	        	flag = TEXTFLAG;
	            stmt = "SELECT textcontent "
	                   "FROM dasl_resource "
	                   "WHERE serialno=%d";
	        }
	        else if ( db_r->istext == 2 )   /*It is a binary file */
	        {
	        	flag = BINFLAG;
	            stmt = "SELECT bincontent "
	                   "FROM dasl_resource "
	                   "WHERE serialno=%d";
	        }
	        else
	        { /* NONE */
	            DBG0 ( "istext is not 1 or 2." );
	            return 100;
	        }
	
	        rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, stmt, 
			                     NULL, 
			                     &pselect);
				
			if(rv != 0) {   	
			   	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
				return -1;
			}
			
			/*Execute the select command */
			rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1,
					apr_ltoa(pool, db_r->serialno) );
					
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
				return -1;
			}
			
	    }

	    /* create temporary file */
    	code = apr_file_open(&file, filename, flag, APR_OS_DEFAULT, pool);

    	
    	/* read content form db into local variable */
	    for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) 
	    {
	    	data = apr_dbd_get_entry(dbd->driver, row, 0);
		}
	    
	    if (db_r->istext == 1 && code == APR_SUCCESS) /*It is a regular file */
    	{
	    	/* write dbd data 1:1 in a file */
    		code = apr_file_write_full(file, data, strlen(data), &bytes);
		} 
    	else if (db_r->istext == 2 && code == APR_SUCCESS) /*It is a binary file */
    	{
    		/* decode BASE64 coded binary data and write in a file */
    		bytes = apr_base64_decode_len(data);
			bindata = apr_palloc(pool, bytes);
			rv = apr_base64_decode_binary(bindata, data);
			code = apr_file_write_full(file, bindata, rv, NULL);
		}
	    
	    code = apr_file_close(file);
	    
	    DBG0 ( "Before return read_content" );
	    return 0;
    }
    else
    {
    	return -1;
    }
}               /*End of dbms_read_content */

/**
 * Write the text content to the dasl_resource table by serialno
 * @param d DB connection struct containing the user, password, and DB name
 * @param bd_r Identifies the resource by bd_r->serialno
 * @return 0 on success, -1 on failure
 */
int dbms_write_content ( dav_repos_db * d,
                         dav_repos_resource * db_r, const char *tpfilename )
{
    apr_pool_t *pool = db_r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	apr_status_t code;
	apr_file_t *file = NULL;
	int rv = 0, nrows = 0, buflen = 0;
	apr_size_t bytes = 0;
	unsigned char *buf = NULL;
	char *cbuf = NULL;
	/* EXPERIMENTAL END */
    
    TRACE();

    DBG1 ( "getcontenttype: %s", db_r->getcontenttype );
    DBG1 ( "filename: %s", tpfilename );
    
    dbd = dbms_get_db(d);
    
	
	/* Read the file into memory */
    code = apr_file_open(&file, tpfilename, APR_READ | APR_BINARY, 0, pool);
    bytes = db_r->getcontentlength;
	buf = apr_palloc(pool, (db_r->getcontentlength+1)*sizeof(unsigned char));
	code = apr_file_read(file, buf, &bytes);
	buf[bytes] = '\0';
	
    if (dbd && code == APR_SUCCESS)
    {

	    /*is this a text or not? */
	    if ( strncasecmp ( db_r->getcontenttype, "text", 4 ) == 0 )
	    {
	    	
	    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "UPDATE dasl_resource \
	                             SET istext=1, textcontent=%s \
	                             WHERE serialno=%d", 
			                     NULL, 
			                     &pquery);
				
			if(rv != 0) {   	
			   	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
				return -1;
			}
			
			/*Execute the update command */
			
		    rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
		    		buf,
					apr_ltoa(pool, db_r->serialno) );
					
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
				return -1;
			}
	    }			
	    else
	    {
			
			/* encode in base64 */
	    	buflen = apr_base64_encode_len(bytes);
	    	cbuf = (char *) apr_palloc(pool, buflen);
	    	rv = apr_base64_encode_binary(cbuf, buf, bytes);
	    	
	    	if(rv != buflen)
	    		return -1;
	    	
	    	
	    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "UPDATE dasl_resource \
	    						 SET istext=2, bincontent=%s \
	    						 WHERE serialno=%d", 
			                     NULL, 
			                     &pquery);
				
			if(rv != 0) {   	
			   	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
				return -1;
			}
			
			/*Execute the update command */
		    rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
		    		cbuf,
					apr_ltoa(pool, db_r->serialno) );
					
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
				return -1;
			}

	    }

	    return 0;
    }
    else
    {
    	return -1;
    }
}               /*End of dbms_write_content */


/**
 * Checks a resource by its serialno whether it is in the resource link or not if exists, set locate_resource to the pointer of the resource whose serialno is 'serialno'
 * @param bd_r A linked list of resources to check
 * @param serialno The serial number of the resource to look for
 * @param resource_pointer Out variable which points to the resource, if found
 * @return 1 if the resource exists, otherwise 0
 */
static int resource_in_link ( dav_repos_resource * db_r, long serialno,
                              dav_repos_resource ** locate_resource )
{
    dav_repos_resource *resource_pointer = NULL;
    resource_pointer = db_r;

    TRACE();

    while ( resource_pointer != NULL )
    {
        if ( resource_pointer->serialno == serialno )
        {
            *locate_resource = resource_pointer;
            return 1;
        }

        resource_pointer = resource_pointer->next;
    }

    *locate_resource = NULL;
    return 0;
}               /*End of resource_in_link */

/**
 * Inserts a dead property struct into the dead property list of the current resource node, if name does exist in the dead property list
 * @param db_r The resource to add the property to
 * @param resource_node ???
 * @param name The name of the property to add
 * @param value The value of the property to add
 *//*
static void add_dead_property_inlink ( dav_repos_resource * db_r,
                                       dav_repos_resource * resource_node,
                                       char *name, char *value )
{
    dav_repos_resource *presource_node = resource_node;
    dav_repos_property *pproperty_node = NULL;
*/    /*Is the previous node of pproperty_node */
/*    dav_repos_property *pupper_property = NULL;

    dav_repos_property *pnew_property = NULL;

    TRACE();

    pproperty_node = presource_node->pr;
    pupper_property = presource_node->pr;
*/
    /*check whether the property with the name 'name' exists
     *in the property link or not*/
/*
    while ( ( pproperty_node != NULL ) &&
            ( apr_strnatcmp ( pproperty_node->name, name ) != 0 ) )
    {
        pupper_property = pproperty_node;
        pproperty_node = pproperty_node->next;
    }

    if ( pproperty_node == NULL )
    {
*/        /*'name' does not exist, so add a new property node */
/*        pnew_property = apr_pcalloc ( db_r->p, sizeof ( *pnew_property ) );
        pnew_property->name = apr_pstrndup ( db_r->p, name, sizeof ( name ) );
        pnew_property->value = apr_pstrndup ( db_r->p, value, sizeof ( value ) );
        pnew_property->next = NULL;

        if ( presource_node->pr == NULL )   *//*Property link is originally empty */
/*            presource_node->pr = pnew_property;
        else            *//*Add the new node to the tail of the link */
/*            pupper_property->next = pnew_property;
    }               *//*End of if (pproperty_node == NULL ) */
/*}               *//*End of add_dead_property_inlink */

/**
 * Execute the search query and return the search result in the link db_r including property link in db_r
 * @param d DB connection struct containing the user, password, and DB name
 * @param search_command SQL query to execute, the serialno field must always appear in fields of the search_command
 * @param db_r Out variable containing the result of the search. The result starts from db_r->next
 * @return 0 on success, -1 on failure
 */
int dbms_search ( dav_repos_db * d, char *search_command,
                  dav_repos_resource * db_r )
{

	apr_pool_t *pool = db_r->p;
    
    /*Pointer to the tail of the resource link */
    struct dav_repos_resource *presult_link_tail = NULL;
    /*Pointer to the new item of resource */
    struct dav_repos_resource *pnew_link_item = NULL;
    /*Pointer to resource in the link currently being set */
    struct dav_repos_resource *presource_current_set = NULL;


    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	/*apr_file_t *file = NULL;
	apr_size_t size = 0;*/
	/* EXPERIMENTAL END */
    
    TRACE();

    DBG1 ( "SEARCH: %s\n", search_command );

/*    size = strlen(search_command);
    apr_file_open(&file, "/tmp/sqlquery", APR_WRITE | APR_CREATE | APR_BUFFERED |APR_TRUNCATE, APR_OS_DEFAULT, db_r->p);
    apr_file_write(file, search_command, &size);
    apr_file_close(file);
  */  
    /*Execute sequry */

    dbd = dbms_get_db(d);
    

    presult_link_tail = db_r;
    presource_current_set = db_r;
    presult_link_tail->next = NULL;
    presult_link_tail->pr = NULL;

    if(dbd)
    {
    
    	apr_dbd_select(dbd->driver, pool, dbd->handle, &res, search_command, 1);
    	
	    /*Deal with every records from the search result */
	    for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1))
	    {   /*fetch every row */
	        long resource_serialno = 0; /*serialno of the resource in row */
	        int resource_exists = 0;    /*resource exists in the resource link */
	
	        /*Check where resource has existed in the resource link */
	        resource_serialno = atol ( apr_dbd_get_entry(dbd->driver, row, 0) );
	        resource_exists =
	            resource_in_link ( db_r->next, resource_serialno,
	                               &presource_current_set );
	
	        if ( resource_exists == 0 )
	        {   /*Resource not in the resource link */
	            /*Resource not the resource link then create a new struct node */
	            /* Malloc a new structure */
	
	            pnew_link_item = dav_repos_mk_blank_resource ( db_r->p );
	            presult_link_tail->next = pnew_link_item;
	            /* Copy pool */
	            pnew_link_item->p = db_r->p;
	            presult_link_tail = pnew_link_item;
	
	            /*Set live property values of every record */
	
	                presult_link_tail->serialno = atol ( apr_dbd_get_entry(dbd->driver, row, 0) );
	                presult_link_tail->uri = apr_pstrdup ( db_r->p, apr_dbd_get_entry(dbd->driver, row, 1) );
	                presult_link_tail->creationdate = atoll ( apr_dbd_get_entry(dbd->driver, row, 2) );
	                presult_link_tail->displayname = apr_pstrdup ( db_r->p, apr_dbd_get_entry(dbd->driver, row, 3) );
	                presult_link_tail->getcontentlanguage = apr_pstrdup ( db_r->p, apr_dbd_get_entry(dbd->driver, row, 4) );
	                presult_link_tail->getcontenttype = apr_pstrdup ( db_r->p, apr_dbd_get_entry(dbd->driver, row, 5) );
	                presult_link_tail->getcontentlength = atol ( apr_dbd_get_entry(dbd->driver, row, 6) );
	                presult_link_tail->getetag = apr_pstrdup ( db_r->p, apr_dbd_get_entry(dbd->driver, row, 7) );
	                presult_link_tail->getlastmodified = atoll ( apr_dbd_get_entry(dbd->driver, row, 8) );
	                presult_link_tail->resourcetype = atoi ( apr_dbd_get_entry(dbd->driver, row, 9) );
	                presult_link_tail->source = atoi ( apr_dbd_get_entry(dbd->driver, row, 10) );
	                /*presult_link_tail->depth = atoi ( apr_dbd_get_entry(dbd->driver, row, 11) );
	                presult_link_tail->istext = atoi ( apr_dbd_get_entry(dbd->driver, row, 12) );*/
	
	            /*Used when setting dead properties */
	            presource_current_set = presult_link_tail;
	        }           /*End of if resource_in_link == 0 */
	
	        presult_link_tail->next = NULL;
	        presult_link_tail->pr = NULL;
	    }               /*End of for() */
	
	    presult_link_tail = db_r;
	    while ( presult_link_tail != NULL )
	    {
	        presult_link_tail = presult_link_tail->next;
	    }
	
	    return 0;
    }
    else
    {
    	return -1;
    }
}               /*End of dbms_search */


/**
 * Removes a resource from the repository. If the resource is collection, all the resources it contains will also be removed.
 * @param d DB connection struct containing the user, password, and DB name.
 * @param r A pointer to the resource to be removed.
 * @return 0 on success, -1 on failure
 */
int dbms_remove_resource ( dav_repos_db * d, const dav_repos_resource * r )
{
    apr_pool_t *pool = r->p;
    char *path = NULL;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	int nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    /* You can't delete it directly */
    if ( r->type == DAV_RESOURCE_TYPE_VERSION )
        return -1;

    /* EXPERIMENTAL START */
	
    dbd = dbms_get_db(d);
    
    
	/* EXPERIMENTAL END */
    
    if ( r->type == DAV_RESOURCE_TYPE_HISTORY )
    { /* Version history */
    	
    	/* EXPERIMENTAL START */
    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "DELETE FROM version_resource "
                       "WHERE serialno=%s", 
                       NULL, 
                       &pquery);
    	    		 
		if (rv != 0) {
			db_error_message_new(dbd, rv, "prepare statement failed", __func__);
			return -1;
		}
		
    	
		rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, apr_ltoa(pool, r->serialno));
		
		if(rv) {
			db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
			return -1;
		}
		
		/* EXPERIMENTAL END */
    }
    else
    { /* Normal resource */
        if ( r->resourcetype == 0 )
        {   /*Is a regular file, not directory */
            
            /* EXPERIMENTAL START */
            rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "DELETE FROM dasl_resource \
                           WHERE serialno=%s", 
                           NULL, 
                           &pquery);
    		 
    		if (rv != 0) {
    			db_error_message_new(dbd, rv, "prepare statement failed", __func__);
    			return -1;
    		}
			/* EXPERIMENTAL END */
            
            /* delete the file-content on harddisk if it is external resource */
            if ( ( r->isexternal==1 ) && ( !generate_path ( &path, r->p, d->file_dir, r->serialno, r->version ) ) )
            {
                apr_file_remove ( path, r->p );
                char *deldir=NULL;
                deldir = apr_psprintf ( pool, "%s/%ld",d->file_dir, r->serialno );
                DBG1 ( "delete directory %s", deldir );
                apr_dir_remove ( deldir, r->p );
            }
            
            rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, apr_ltoa(pool, r->serialno));
    		
    		if(rv) {
    			db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
    			return -1;
    		}
            
        }               /*End of if it is a regular file */
        else if ( r->resourcetype == 1 )
        {   /*It is a collection(directory) */
            /* first delete harddisk files */
            
        	/* EXPERIMENTAL START */
			rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT serialno \
                               FROM dasl_resource \
                               WHERE uri LIKE %s \
                               AND isexternal=1", 
                               NULL, 
                               &pselect);
			
			if (rv != 0) {
				db_error_message_new(dbd, rv, "prepare statement failed", __func__);
    			return -1;
    		}
			
			rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res,  pselect, 1, 
					apr_psprintf ( pool, "%s/%%", r->uri ));
			
			if(rv) {
				db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
				return -1;
			}
			
			for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
				rv != -1;
				rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) 
			{
		
			/* EXPERIMENTAL END */
        	
                generate_path ( &path, pool, d->file_dir, atol(apr_dbd_get_entry(dbd->driver, row, 0)), r->version );
                apr_file_remove ( path, pool );
                char *deldir=NULL;
                deldir = apr_psprintf( pool, "%s/%ld", d->file_dir, atol( apr_dbd_get_entry(dbd->driver, row, 0)) );
                DBG1 ( "delete directory %s", deldir );
                apr_dir_remove ( deldir, r->p );
            }
            
			/* EXPERIMENTAL START */
			/* now delete database content */
			
			rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "DELETE FROM dasl_resource \
	                   WHERE uri=%s OR uri LIKE %s",
	                   NULL, 
	                   &pquery);
			
			if (rv != 0) {
				db_error_message_new(dbd, rv, "prepare statement failed", __func__);
    			return -1;
    		}
			
			rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, 
			                   &nrows, 
			                   pquery,
			                   r->uri,
			                   apr_psprintf ( pool, "%s/%%", r->uri ));
			
			if(rv) {
				db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
				return -1;
			}
			
			
			/* EXPERIMENTAL END */
            
            
            
        }
        else
        { /* huh? */
            return -1;
        }
    }

    /* Set the version things in the 'dasl_resource' */
    if ( r->type == DAV_RESOURCE_TYPE_HISTORY )
    { /* Version property */
        
        /* EXPERIMENTAL START */
		rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "UPDATE dasl_resource "
                           "SET isversioned = 0, checkin=-1, checkout = -1 "
                           "WHERE serialno=%d", 
                           NULL, 
                           &pquery);
		
		if (rv != 0) {
			db_error_message_new(dbd, rv, "prepare statement failed", __func__);
			return -1;
		}
		
		rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, 
		                   &nrows, 
		                   pquery,
		                   apr_ltoa(pool, r->serialno));
		
		if(rv) {
			db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
			return -1;
		}
		
		/* EXPERIMENTAL END */
        
    }

    return 0;
}               /*End of dbms_remove_resource */

/**
 * Adds a new dead property record the the dead property table in the DB
 * @param d DB connection struct
 * @param r Identifies the resource to associate the property with
 * @param pr The property to insert
 * @return 0 on success, -1 on failure
 */
int db_insert_property ( dav_repos_db * d,
                         const dav_repos_resource * r,
                         const dav_repos_property * pr )
{
    apr_pool_t *pool = r->p;

    /* EXPERIMENTAL START */
    ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	int rv = 0;
	int nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();
       
    /* EXPERIMENTAL START */
	dbd = dbms_get_db(d);
	
	if (dbd) {
		/*Create the insert command */
		rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "INSERT INTO dasl_property "
                       "(serialno, ns_id, name, value) "
                       "VALUES(%d, %d, %s, %s)", 
                       NULL, &pquery);
		
		if(rv != 0) {
			db_error_message_new(dbd, rv, "prepare statement failed", __func__);
			return rv;
		}
		
		rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery,
				apr_ltoa(pool, r->serialno),
				apr_ltoa(pool, pr->ns_id),
				dav_repos_dbms_string(pr->name),
				dav_repos_dbms_string(pr->value));
		
		if(rv != 0) {
			db_error_message_new(dbd, rv, "Insert property failed", __func__);
			return rv;
		}
		
	}
	/* EXPERIMENTAL END */
    
	return 0;
}               /*End of db_insert_property */

/**
 * Set the value of a dead property on a resource identified by serialno
 * @param d DB connection struct
 * @param r Identified the resource to set the property on
 * @param pr The dead property to set
 * @return 0 on success, -1 on failure
 */
int dbms_set_dead_property ( dav_repos_db * d,
                             const dav_repos_resource * r,
                             const dav_repos_property * pr )
{
    apr_pool_t *pool = r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	int nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    /* EXPERIMENTAL START */
	dbd = dbms_get_db(d);
	
	if (dbd) {
		
		rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT serialno FROM dasl_property"
                       " WHERE serialno=%d and name=%s and ns_id=%d", 
                       NULL, 
                       &pselect);
		
		if(rv != 0) {
			db_error_message_new(dbd, rv, "prepare statement failed", __func__);
			return rv;
		}
		
		rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res,  pselect, 1, 
							apr_ltoa(pool, r->serialno),
							pr->name,
							apr_ltoa(pool, pr->ns_id));
					
		if(rv != 0) {
			db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
			return -1;
		}
		
		nrows = apr_dbd_num_tuples(dbd->driver, res);
		
		rv = apr_dbd_get_row(dbd->driver, pool, res, &row, nrows+1);
		
		
		if ( nrows == 0 ) {
			/*Insert a new property record */
			return db_insert_property ( d, r, pr );
		}
		
		nrows = 0;
		
		rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "UPDATE dasl_property \
                       SET ns_id=%d, name=%s, value=%s \
                       WHERE serialno=%d and name=%s", 
                       NULL, 
                       &pquery);
		
		if(rv != 0) {
			db_error_message_new(dbd, rv, "prepare statement failed", __func__);
			return rv;
		}
		
		rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
				apr_ltoa(pool, pr->ns_id),
				pr->name,
				dav_repos_dbms_string(pr->value),
				apr_ltoa(pool, r->serialno),
				pr->name );
		
		if(rv != 0) {
			db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
			return -1;
		}
		
		return 0;
	}
	else
	{
		return -1;
	}
	/* EXPERIMENTAL END */
}               /*End of dbms_set_dead_property */


/**
 *  I'll fill pr->name , r->serialno and you will fill pr->value
 * @param d DB connection struct
 * @param r Identifies the resource
 * @param pr Identifies the property to retrieve
 * @return 0 on success, 100 if not found, -1 on failure
 * */
int dbms_get_dead_property ( dav_repos_db * d,
                             const dav_repos_resource * r,
                             dav_repos_property * pr )
{
    apr_pool_t *pool = r->p;
    const char *value;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    dbd = dbms_get_db(d);
    
    if(dbd)
    {
	    /*Create the search command */
	    if ( r->type == DAV_RESOURCE_TYPE_VERSION )
	    { /* Version_property */
	        
	    	/* EXPERIMENTAL START */
			rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT ns_id, value FROM version_property "
	                           "Where serialno=%d and version=%d and name=%s and ns_id=%d", 
	                           NULL, 
	                           &pselect);
			
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepare statement failed", __func__);
				return -1;
			}
			
			rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1, 
					apr_ltoa(pool, r->serialno),
					apr_itoa(pool, r->version),
					pr->name,
					apr_ltoa(pool, pr->ns_id));
			
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
				return -1;
			}

			/* EXPERIMENTAL END */
	    }
	    else
	    { /* Normal property */
	        
	    	/* EXPERIMENTAL START */
			rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT ns_id, value FROM dasl_property "
	                           "Where serialno=%d and name=%s and ns_id=%d", 
	                           NULL, 
	                           &pselect);
			
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepare statement failed", __func__);
				return -1;
			}
			
			rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1, 
					apr_ltoa(pool, r->serialno),
					pr->name,
					apr_ltoa(pool, pr->ns_id));
			
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
				return -1;
			}

			/* EXPERIMENTAL END */
	    }
	    
	    /* EXPERIMENTAL START */
		for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) 
		{
			pr->ns_id = atol(apr_dbd_get_entry(dbd->driver, row, 0));
			value = apr_dbd_get_entry(dbd->driver, row, 1);
			pr->next = NULL;
		}
		/* EXPERIMENTAL END */

	    return 0;
    }
    else
    	return -1;
}               /*End of dbms_dead_property */

/**
 * Moves a resource. If the source is a regular file,
 * just change the URI. If the source is collection,
 * move all files in the collection
 * @param d DB connection struct
 * @param r_src source
 * @param r_des destination ( only contains URI field )
 * @return 0 on success -1 on failure
 */
int dbms_move_resource ( dav_repos_db * d,
                         const dav_repos_resource * r_src,
                         const dav_repos_resource * r_des )
{
    int des_depth = 0;      /* Depth of the destination URI */
    int depth_diff = 0;
    apr_pool_t *pool = r_src->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	int rv = 0;
	int nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    des_depth = ap_count_dirs ( r_des->uri );
    depth_diff = des_depth - ap_count_dirs ( r_src->uri );

    dbd = dbms_get_db(d);
    
    if (dbd) 
    {
		
	    if ( r_src->resourcetype == 0 )
	    {
	        /*Is a regular file, not directory
	         *Change the URI directly*/
	        
	    	/* EXPERIMENTAL START */
			rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "UPDATE dasl_resource \
		                         SET URI=%s, \
		                         displayname=%s, \
		                         depth=%d \
		                         WHERE serialno=%d", 
					             NULL, 
					             &pquery);
					
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepare statement failed", __func__);
				return -1;
			}
				
			rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
								 r_des->uri,
								 apr_pstrdup ( pool, basename( r_des->uri ) ),
								 apr_itoa(pool, des_depth ),
								 apr_ltoa(pool, r_src->serialno));
				
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
		  		return -1;
			}
				
			
			/* EXPERIMENTAL END */
	    	
	    }               /*End of if it is a regular file */
	    else if ( r_src->resourcetype == 1 )
	    {  /*It is a collection(directory) */
	        
	    	/* EXPERIMENTAL START */
			rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "UPDATE dasl_resource "
	                             "SET URI = CONCAT_SUBSTRING(%s, URI, %d), "
	                             "depth=depth + %d "
	                             "WHERE URI = %s OR URI LIKE %s ", 
			                     NULL, 
			                     &pquery);
				
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepare statement failed", __func__);
				return -1;
			}
					
			rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
					r_des->uri,
					apr_itoa( pool, strlen(r_src->uri ) + 1),
					apr_itoa( pool, depth_diff),
					r_src->uri,
					apr_psprintf( pool, "%s/%%", r_src->uri ));
					
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
				return -1;
			}
			
			/* EXPERIMENTAL END */
	    	
	    }
	    else
	    { /* huh? */
	        return -1;
	    }
    } 
    else 
    { /* no db connection established */
		return -1;
	}

    return 0;
}               /*End of dbms_move_resource */

/**
 * Copy the source to the destination. If the source is a regular file,
 * just copy the record in the database with the new URI.
 * If the source is collection, duplicate all records in the collection
 * in the database with the new URIs
 * @param d DB connection struct
 * @param r_src The source
 * @param r_des The destination ( only contains URI field )
 * @return 0 on success, -1 on failure
 */
int dbms_copy_resource ( dav_repos_db * d,
                         const dav_repos_resource * r_src,
                         const dav_repos_resource * r_des,
                         int depth )
{
    int des_depth = 0;      /* Depth of the destination URI */
    int depth_diff = 0;
    apr_pool_t *pool = r_src->p;
    char *src_fn = NULL, *des_fn = NULL;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_results_t *res2 = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	int nrows = 0;
	int error = 0;
	const char *message;
	/* EXPERIMENTAL END */
    
    TRACE();

    dbd = dbms_get_db(d);
    
    if(dbd) 
    {
	    des_depth = ap_count_dirs ( r_des->uri );
	    depth_diff = des_depth - ap_count_dirs ( r_src->uri );
	
	    /* Drop the temporary table first */
	    rv = apr_dbd_query(dbd->driver, dbd->handle, &nrows, "DROP TABLE tmp_resrecord");
	    
	    if(rv != 0) {
		  	message = apr_dbd_error(dbd->driver, dbd->handle, rv);
		  	ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, d->r->server, 
		  			"prepared statement execution failed: NOTICE %d, Message: %s", rv, message);
		}

	    if ( r_src->resourcetype == 0 || depth == 0 )
	    {
	        /*Source Is a regular file, not directory */
	        /*Copy the record in the database with the new URI */
	        /*Store the source record into a temporary table */
	    	
	    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT CREATE_TEMPORARY_RESOURCE_TABLE( %s, %s, %d, %d)", NULL, &pselect);
	    	
	    	if(rv != 0) {
	    		db_error_message_new(dbd, rv, "prepare statement failed", __func__);
				return -1;
			}
	    	
	    	rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1,
	    			r_des->uri,
	    			apr_pstrdup(pool, basename(r_des->uri) ),
	    			apr_itoa(pool, des_depth),
	    			apr_ltoa(pool, r_src->serialno));
	    	
	    	if(rv != 0) {
	    		db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
				return -1;
			}
	    	
	    	for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
				rv != -1;
				rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
				
			}
	    	
	    	
	        /* Duplicate the records in the temporary table
	           to the table dasl_resource */
	    	
	    	rv = apr_dbd_query(dbd->driver, dbd->handle, &nrows, "INSERT INTO dasl_resource "
	                           "SELECT * FROM tmp_resrecord");
	    	
	    	if(rv != 0) {
	    		db_error_message_new(dbd, rv, "statement execution failed", __func__);
				return -1;
			}

	        /* now copy the dead properties *
	         * TODO return value should be checked
	         */
	    	
	        dbms_copy_dead_properties ( d, r_src, r_des );
	
	        if ( r_src->isexternal && r_src->resourcetype==0 )
	        {
	            /*long des_serialno = mysql_insert_id ( ( MYSQL * ) ( & ( d->db ) ) );*/
	        	long des_serialno = dbms_insert_id(d, "dasl_resource", "serialno");
	            DBG0 ( "file is externally stored, copying..." );
	            if ( generate_path ( &src_fn, r_src->p, d->file_dir,
	                                 r_src->serialno, r_src->version ) )
	            {
	
	            	ap_log_error ( APLOG_MARK, APLOG_ERR, 0, NULL,
	                                   "Error generating source path." );
	                return -1;
	            }
	            // new resource is version 0 by default
	            if ( generate_path ( &des_fn, r_src->p, d->file_dir,
	                                 des_serialno, 0 ) )
	            {
	            	ap_log_error ( APLOG_MARK, APLOG_ERR, 0, NULL,
	                                   "Error generating destination path." );
	                return -1;
	            }
	            DBG2 ( "copying file %s to %s", src_fn, des_fn );
	            if ( apr_file_copy ( src_fn, des_fn, APR_OS_DEFAULT, r_src->p )
	                    != APR_SUCCESS )
	            {
	            	ap_log_error ( APLOG_MARK, APLOG_ERR, 0, NULL, 
	            					"Unable to copy file" );
	                return -1;
	            }
	        }
	    }               /*End of if it is a regular file */
	    else if ( r_src->resourcetype == 1 )
	    {
	        /*It is a collection(directory) */
	        /* duplicate the records in the collection
	           in the database with the new URIs */
	        /*Store the source records into a temporary table */
	        
	    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT CREATE_TEMPORARY_COLL_RESOURCE_TABLE( %s, %s, %d, %d)", NULL, &pselect);
	    		    	
	    	if(rv != 0) {
	    		db_error_message_new(dbd, rv, "prepare statement failed", __func__);
				return -1;
			}
	    	
	    	rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1,
	    			r_des->uri,
	    			r_src->uri,
	    			apr_itoa(pool, strlen ( r_src->uri ) + 1),
	    			apr_itoa(pool, depth_diff));
	    	
	    	if(rv != 0) {
	    		db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
				return -1;
			}
	    	
	    	for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
				rv != -1;
				rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
				
			}
	
	        /* Duplicate the records in the temporary table
	           to the table dasl_resource */
	    	
	    	rv = apr_dbd_query(dbd->driver, dbd->handle, &nrows, "INSERT INTO dasl_resource "
	    		                           "SELECT * FROM tmp_resrecord");
	    		    	
	    	if(rv != 0) {
                db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
				return -1;
			}
	
	        /* copy the dead properties
	         * TODO return value should be checked
	         */
	        dbms_copy_dead_properties ( d, r_src, r_des );
	        
	        /*const char *stmt = apr_psprintf ( pool, "SELECT s.serialno, d.serialno "
    		"FROM dasl_resource s, dasl_resource d "
    		"WHERE s.uri LIKE '%s/%%' AND "
    		"d.uri LIKE '%s/%%' AND "
    		"s.resourcetype = 0 AND "
    		"d.resourcetype = 0 AND " 
    		"s.isexternal != 0 AND "
    		"d.isexternal != 0 AND " 
    		"SUBSTRING(s.uri, %d) = SUBSTRING(d.uri, %d)", 
    		r_src->uri, 
    		r_des->uri, 
    		(strlen ( r_src->uri ) + 1),
    		(strlen ( r_des->uri ) + 1));

	        rv = apr_dbd_select(dbd->driver, pool, dbd->handle, &res, stmt, 0);*/
	        
	        rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT s.serialno, d.serialno "
	        		"FROM dasl_resource s, dasl_resource d "
	        		"WHERE s.uri LIKE %s AND "
	        		"d.uri LIKE %s AND "
	        		"s.resourcetype = 0 AND "
	        		"d.resourcetype = 0 AND " 
	        		"s.isexternal != 0 AND "
	        		"d.isexternal != 0 AND " 
	        		"SUBSTRING(s.uri, %d) = SUBSTRING(d.uri, %d)", NULL, &pselect);
	        
	        if(rv != 0) {
	        	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
				return -1;
			}
	        
	        res2 = NULL;
	        rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res2, pselect, 1,
	        		apr_psprintf ( pool, "%s/%%", r_src->uri ),
	        		apr_psprintf ( pool, "%s/%%", r_des->uri ),
	        		apr_itoa(pool, (strlen ( r_src->uri ) + 1)),
	        		apr_itoa(pool, (strlen ( r_des->uri ) + 1)));
	        
	        if(rv != 0) {
	        	db_error_message_new(dbd, rv, "statement execution failed", __func__);
				return -1;
			}
	        
	        
	        for(rv = apr_dbd_get_row(dbd->driver, pool, res2, &row, -1);
				rv != -1;
				rv = apr_dbd_get_row(dbd->driver, pool, res2, &row, -1)) 
			{
	            char *src_fn = NULL, *des_fn = NULL;
	            
	            long src_serialno = atol(apr_dbd_get_entry(dbd->driver, row, 0));
	            long des_serialno = atol(apr_dbd_get_entry(dbd->driver, row, 1));
	
	            if ( generate_path ( &src_fn, r_src->p, d->file_dir,
	                                 src_serialno, 0 ) )
	            {
	                db_error_message_new (dbd, rv, "Error generating source path.", __func__);
	                error = 1;
	            }
	
	            if ( generate_path ( &des_fn, r_src->p, d->file_dir,
	                                 des_serialno, 0 ) )
	            {
	                db_error_message_new (dbd, rv, "Error generating destination path.", __func__);
	                error = 1;
	            }
	
	            if ( apr_file_copy ( src_fn, des_fn, APR_OS_DEFAULT, r_src->p )
	                    != APR_SUCCESS )
	            {
	                db_error_message_new (dbd, rv, "Unable to copy file" , __func__);
	                error = 1;
	            }
	        }
	        
	        if (error) {
				return -1;
			}
	        
	    }               /*End of else if */
	
	    return 0;
    }
    else /* No DB connection */
    {
    	return -1;
    }
}               /*End of dbms_copy_resource */


/**
 * Copy all dead properties from the source to the destination resource.
 * If the source is a regular file, just copy properties of one resource.
 * If the source is collection, all properties  will be recursivly copied.
 * @param d DB connection struct
 * @param r_src The source
 * @param r_des The destination ( only contains URI field )
 * @return 0 on success, -1 on failure
 */
int dbms_copy_dead_properties ( dav_repos_db * d,
                                const dav_repos_resource * r_src,
                                const dav_repos_resource * r_des )
{
    apr_pool_t *pool = r_src->p;
    int i=0, numcopy=0;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	int nrows = 0;
	const char *message;
	/* EXPERIMENTAL END */
    
    TRACE();

    dbd = dbms_get_db(d);
    
    if (dbd) {

	    // copy dead properties of all resources
	    rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT s.serialno, \
	                       d.serialno \
	                       FROM dasl_resource s, dasl_resource d \
	                       WHERE s.uri LIKE %s AND d.uri LIKE %s \
	                       AND SUBSTRING(s.uri, %d) = SUBSTRING(d.uri, %d)", 
	                       NULL, 
	                       &pselect);
	    
	    if(rv != 0) {
    		db_error_message_new ( dbd, rv, "Error while prepareing command" , __func__);
    		return -1;
    	}
	    
	    rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1, 
	    		apr_psprintf ( r_src->p, "%s%%", r_src->uri ),
	    		apr_psprintf ( r_src->p, "%s%%", r_des->uri ),
	    		apr_itoa(pool, strlen ( r_src->uri ) + 1),
	    		apr_itoa(pool, strlen ( r_des->uri ) + 1));
	    
	    if(rv != 0) {
    		db_error_message_new ( dbd, rv, "Error while execution command" , __func__);
    		return -1;
    	}

	    numcopy = apr_dbd_num_tuples(dbd->driver, res);
	    
	    long *src_ptr = apr_pcalloc ( pool, 2 * ( numcopy+1 ) * sizeof ( long ) );
	    long *des_ptr = apr_pcalloc ( pool, 2 * ( numcopy+1 ) * sizeof ( long ) );
	   
	    for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) 			
	    {
	        src_ptr[i] = atoi(apr_dbd_get_entry(dbd->driver, row, 0));
	        des_ptr[i] = atoi(apr_dbd_get_entry(dbd->driver, row, 1));
	        i++;
	    }

	
	    for ( i=0; i<=numcopy; i++ )
	    {
	
	        /* Drop the temporary table */
	    	rv = apr_dbd_query(dbd->driver, dbd->handle, &nrows, "DROP TABLE tmp_resprops");
		    
		    if(rv != 0) {
			  	message = apr_dbd_error(dbd->driver, dbd->handle, rv);
			  	ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, d->r->server, 
			  			"DROP TABLE tmp_resprops failed: NOTICE %d, Message: %s", rv, message);
			}
	    	
		    rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT CREATE_TEMPORARY_PROPERTY_TABLE(%d, %d)", NULL, &pselect);
	
		    if(rv != 0) {
        		db_error_message_new ( dbd, rv, "Error while preparing command" , __func__);
        		return -1;
        	}
		    
		    res = NULL;
		    /*Execute the command */
		    rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1, 
		    		apr_ltoa(pool, des_ptr[i]),
		    		apr_ltoa(pool, src_ptr[i]));
		    
		    if(rv != 0) {
        		db_error_message_new ( dbd, rv, "Error while execution command" , __func__);
        		return -1;
        	}
		    
		    for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
				rv != -1;
				rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
				
			}
	        
		    rv = apr_dbd_query(dbd->driver, dbd->handle, &nrows, "INSERT INTO dasl_property "
	                           "SELECT * FROM tmp_resprops");
		    
		    if(rv != 0) {
        		db_error_message_new ( dbd, rv, "Error in the insert command" , __func__);
        		return -1;
        	}
		    
	    }
	    return  0;
	    
    }
    else
    { /* No DB connection */
    	return -1;
    }
}

/**
 * Exists check using URI
 * @param d DB connection struct
 * @param r A pointer to a resource struct containing the memory pool to use
 * @param uri The URI of the resource to check
 * Return 0: Not Found 1: Found -1: Failure
 */
int dbms_exists ( dav_repos_db *db, const char *uri, apr_pool_t *pool )
{

    TRACE();

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0, numrows = 0;
	/* EXPERIMENTAL END */
    
    /*Create the search command */
    dbd = dbms_get_db(db);
	
    if (dbd) 
    {
		rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT serialno "
                       "FROM dasl_resource "
                       "WHERE uri=%s", NULL, &pselect);
		
		if(rv != 0) {
    		db_error_message_new ( dbd, rv, "Error while preparing command" , __func__);
    		return -1;
    	}
		
		rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1, uri);
		
		if(rv != 0) {
    		db_error_message_new ( dbd, rv, "Error while execution command" , __func__);
    		return -1;
    	}
		
		numrows = apr_dbd_num_tuples(dbd->driver, res);
		
		for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
			
		}
		
		if (numrows > 0)
			return 1;
		else
			return 0;
		
	} 
    else 
    {
		return -1;
	}
}               /*End of dbms_exists */


/**
 * Get properties in a form of property link for every resource
 * @param d DB connection struct
 * @param db_r
 * @return 0 on success, -1 on failure
 */
int dbms_fill_dead_property ( dav_repos_db * d, dav_repos_resource * db_r )
{
    struct dav_repos_resource *presource_link_tail = NULL;
    struct dav_repos_property *pproperty_link_tail = NULL;
    struct dav_repos_property *pnew_link_item = NULL;
    apr_pool_t *pool = db_r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    presource_link_tail = db_r;

    dbd = dbms_get_db(d);
    
    if(dbd)
    {
	    while ( presource_link_tail != NULL )
	    {
	    	
    	/*Create the search command */
			rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT serialno, \
                           ns_id, \
                           name, \
                           value \
                           FROM dasl_property WHERE serialno=%d \
                           ORDER BY name", 
                           NULL, 
                           &pselect);
			
			if(rv != 0) {
	    		db_error_message_new ( dbd, rv, "Error while preparing command" , __func__);
	    		return -1;
	    	}
			
			res = NULL;
			rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1,
					apr_ltoa(pool, presource_link_tail->serialno));
			
			if(rv != 0) {
	    		db_error_message_new ( dbd, rv, "Error while execution command" , __func__);
	    		return -1;
	    	}
			
			for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
				rv != -1;
				rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
				

	            /*Malloc a new property structure */
	            pnew_link_item = apr_pcalloc ( pool, sizeof ( *pnew_link_item ) );

	            if ( presource_link_tail->pr == NULL )
	            {
	                /*It is the first property record */
	                presource_link_tail->pr = pnew_link_item;
	                pproperty_link_tail = pnew_link_item;
	            }
	            else
	            {
	                /*It is not the first property record */
	                if ( pproperty_link_tail )
	                    pproperty_link_tail->next = pnew_link_item;
	                pproperty_link_tail = pnew_link_item;
	            }
	            
	            /*Get the data of every column of this row to a link */
	            
	            pproperty_link_tail->serialno = atol ( apr_dbd_get_entry(dbd->driver, row, 0) );

	            pproperty_link_tail->ns_id = atol ( apr_dbd_get_entry(dbd->driver, row, 1) );

	            pproperty_link_tail->name = apr_pstrdup(pool, apr_dbd_get_entry(dbd->driver, row, 2));
	                /*( dbrow[2] == NULL ) ? NULL : apr_pstrndup ( db_r->p, dbrow[2],
	                        lengths[2] );*/

	            pproperty_link_tail->value = apr_pstrdup(pool, apr_dbd_get_entry(dbd->driver, row, 3));
	                /*( dbrow[3] == NULL ) ? NULL : apr_pstrndup ( db_r->p, dbrow[3],
	                        lengths[3] );*/

	            pproperty_link_tail->next = NULL;
				
			}
			
			presource_link_tail = presource_link_tail->next;    /*Next resource */
			
		}/*End of while presource_link_tail != NULL */
	    
	    return 0;
    }
	else 
	{
		return -1;
	}
                      
}               /*End of db_get_collection_property */

/**
 * Retrieve all the resources in a collection
 * @param d DB connection struct
 * @param db_r A pointer to a resource struct which identifies the collection
 * @param depth==1 one depth, other infinity
 * @return 0 on success, -1 on failure
 */
int dbms_get_collection_resource ( dav_repos_db * d,
                                   dav_repos_resource * db_r, int depth )
{
    struct dav_repos_resource *presult_link_tail = NULL;
    struct dav_repos_resource *pnew_link_item = NULL;
    apr_pool_t *pool = db_r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    dbd = dbms_get_db(d);
    
    if (dbd) 
    {
    	/*Create the search command */
    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT serialno,\
                       uri, \
                       creationdate, \
                       displayname, \
                       getcontentlanguage, \
                       getcontentlength, \
                       getcontenttype, \
                       getetag, \
                       getlastmodified, \
                       resourcetype, \
                       source, \
                       depth, \
                       istext, \
                       isversioned, \
                       checkin, \
                       checkout, \
                       isexternal \
                       FROM dasl_resource WHERE uri LIKE %s  \
                       AND uri <> %s \
                       AND (%d <> 1 OR depth = %d) \
                       ORDER BY resourcetype DESC, uri, serialno ", NULL, &pselect);
    	
    	if(rv != 0) {
    		db_error_message_new ( dbd, rv, "Error while preparing command", __func__ );
    		return -1;
    	}
    	
    	rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1,
    			apr_psprintf(pool, "%s/%%", db_r->uri),
			    db_r->uri,
			    apr_itoa(pool, depth),
			    apr_itoa(pool, db_r->depth + 1));
    	
    	if(rv != 0) {
    		db_error_message_new ( dbd, rv, "Error while execution command" , __func__);
    		return -1;
    	}
	
    	presult_link_tail = db_r;
	    presult_link_tail->next = NULL;
	    presult_link_tail->pr = NULL;
	    
    	
		for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
		
		
		    /*Malloc a new structure */
		    pnew_link_item = apr_pcalloc ( db_r->p, sizeof ( *pnew_link_item ) );
		    presult_link_tail->next = pnew_link_item;
		    presult_link_tail = pnew_link_item;
		
		    /*Get the data of every column of this row to a link */
		
		    presult_link_tail->serialno = atol ( apr_dbd_get_entry(dbd->driver, row, 0) );
		
		    /* For Catacomb-V
		    presult_link_tail->filename =
		    (dbrow[1] == NULL) ? NULL : apr_pstrndup(db_r->p, dbrow[1],
		       lengths[1]);
		    */
		
		    presult_link_tail->uri = apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row, 1) );
		
		    presult_link_tail->creationdate = atoll ( apr_dbd_get_entry(dbd->driver, row, 2) );
		
		    presult_link_tail->displayname = apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row, 3) );
		
		    presult_link_tail->getcontentlanguage = apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row, 4) );
		
		    presult_link_tail->getcontentlength = atol ( apr_dbd_get_entry(dbd->driver, row, 5) );
		    presult_link_tail->getcontenttype = apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row, 6) );
		
		    presult_link_tail->getetag = apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row, 7) );
		
		    presult_link_tail->getlastmodified = atoll ( apr_dbd_get_entry(dbd->driver, row, 8) );
		
		    presult_link_tail->resourcetype = atoi ( apr_dbd_get_entry(dbd->driver, row, 9) );
		
		    presult_link_tail->source = atoi ( apr_dbd_get_entry(dbd->driver, row, 10) );
		
		    presult_link_tail->depth = atoi ( apr_dbd_get_entry(dbd->driver, row, 11) );
		
		    presult_link_tail->istext = atoi ( apr_dbd_get_entry(dbd->driver, row, 12) );
		
		    presult_link_tail->isversioned = atoi ( apr_dbd_get_entry(dbd->driver, row, 13) );
		    presult_link_tail->checkin = atoi ( apr_dbd_get_entry(dbd->driver, row, 14) );
		    presult_link_tail->checkout = atoi ( apr_dbd_get_entry(dbd->driver, row, 15) );
		    presult_link_tail->isexternal = atoi ( apr_dbd_get_entry(dbd->driver, row,16) );
		
		    presult_link_tail->next = NULL;
		    presult_link_tail->pr = NULL;
		}               /*End of for */
		
		return 0;
    
	} 
    else 
    { /* DB connenction failed*/
		return -1;
	}

}               /*End of dbms_get_collection_resource */

/* Deletes a dasl_property by serialno and property name
 * @param d DB connection struct
 * @param r A pointer to a resource struct identifying
 * the resource to be operated upon
 * @param pr A pointer to a property struct identifying the property to delete
 * @return 0 on success, 1 on failure
 * */
int dbms_del_dead_property ( dav_repos_db * d,
                             const dav_repos_resource * r,
                             const dav_repos_property * pr )
{
    apr_pool_t *pool = r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	int rv = 0;
	int nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    /* EXPERIMENTAL START */
	dbd = dbms_get_db(d);
	
	if (dbd) {
		rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "DELETE FROM dasl_property \
                       WHERE serialno=%d and name=%s and ns_id=%d", 
                       NULL, 
                       &pquery);
		
		if (rv != 0) {
			db_error_message_new(dbd, rv, "prepare statement failed", __func__);
			return -1;
		}
		
		rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
				apr_ltoa(pool, r->serialno),
				dav_repos_dbms_string(pr->name),
				apr_ltoa(pool, pr->ns_id));
		
		if(rv != 0) {
			db_error_message_new(dbd, rv, "Delete property failed", __func__);
			return -1;
		}
	}
	/* EXPERIMENTAL END */
    return 0;
}               /*End of dbms_del_dead_property */

/* Get resourcetype by uri
 * @param d DB connection struct
 * Return 0: success; 100: not found; -1: failure
 * */
int dbms_get_resourcetype ( dav_repos_db * d, dav_repos_resource * r )
{
    apr_pool_t *pool = r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0, rows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    dbd = dbms_get_db(d);
    
    if (dbd) 
    {
		
    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT resourcetype FROM dasl_resource \
                       WHERE uri=%s", NULL, &pselect);
    	
    	if (rv != 0) {
	    	db_error_message_new( dbd, rv, "get_resourcetype() - prepare statement failed", __func__ );
			return -1;
		}
    	
    	rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1, 
    			r->uri);
    	
    	if (rv != 0) {
	    	db_error_message_new( dbd, rv, "get_resourcetype() - statement execution failed", __func__ );
			return -1;
		}
    	
    	for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
			rows++;
			
			r->resourcetype = atoi ( apr_dbd_get_entry(dbd->driver, row, 0) );
			
		}
    	
    	if (rows > 0) 
			return 0;
		else 
			return 100;
    	
	} 
    else 
    {
		return -1;
	}
    
}               /*End of dbms_get_resourcetype */

/* Build namespace
 *
 * Read namespace infor from DB and fill hash
 * db_r must have pool
 * This hash is for PROPFIND
 * */
void dbms_build_ns_id_hash ( dav_repos_db * db, dav_repos_resource * db_r )
{
    apr_pool_t *pool = db_r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	char *ns_id = NULL;
	char *name = NULL;
	const char* temp;
    /* EXPERIMENTAL END */
    
    TRACE();

    /* id_hash is not clean */
    if ( db_r->ns_id_hash != NULL )
        return;

    /* Build hash */
    DBG0 ( "NS_id_hash_building" );
    db_r->ns_id_hash = apr_hash_make ( pool );

    /* EXPERIMENTAL START */
	dbd = dbms_get_db(db);
	
	if (dbd) {
		
		rv = apr_dbd_select(dbd->driver, pool, dbd->handle, &res, "SELECT ns_id , name FROM dasl_namespace", 1);
		
		if(rv != 0) {
			db_error_message_new(dbd, rv, "Select Statement failed", __func__);
			return;
		}
		
    	for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
    		
    		if(rv != 0) {
    			db_error_message_new(dbd, rv, "Error while reading data", __func__);
    		}
    		temp = apr_dbd_get_entry(dbd->driver, row, 0 );
		    ns_id = apr_pstrndup(pool, temp, strlen(temp));
		    temp = apr_dbd_get_entry(dbd->driver, row, 1 );
		    name = apr_pstrndup(pool, temp, strlen(temp));
		    
		    if ( ns_id && name )
            {
                apr_hash_set ( db_r->ns_id_hash, ns_id, APR_HASH_KEY_STRING, name );
                DBG2 ( "ADD Hash : %s: %s", ns_id, name );
            }
    	}
		
	}
	/* EXPERIMENTAL END */
	
    return;
}

/* Get namespace
 * If new namespace, save it and return ns_id
 *
 * This hash is for PROPPATCH
 * It must be rebuild after adding new namespace
 * */
static void dbms_build_ns_hash ( dav_repos_db * db,
                                 dav_repos_resource * db_r )
{
    apr_pool_t *pool = db_r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	char *ns_id = NULL;
	char *name = NULL;
	const char* temp;
    /* EXPERIMENTAL END */
    
    TRACE();

    /* Already built */
    if ( db_r->ns_hash )
        return;

    DBG0 ( "NS_hash_building" );
    db_r->ns_hash = apr_hash_make ( db_r->p );

    /* EXPERIMENTAL START */
	dbd = dbms_get_db(db);
	
	if (dbd) {
		
		rv = apr_dbd_select(dbd->driver, pool, dbd->handle, &res, "SELECT ns_id , name FROM dasl_namespace", 1);
		
		if(rv != 0) {
			db_error_message_new(dbd, rv, "Select Statement failed", __func__);
			return;
		}
		

    	for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
    		
    		if(rv != 0) {
    			db_error_message_new(dbd, rv, "Error while reading data", __func__);
    		}

		    temp = apr_dbd_get_entry(dbd->driver, row, 0 );
		    ns_id = apr_pstrndup(pool, temp, strlen(temp));
		    temp = apr_dbd_get_entry(dbd->driver, row, 1 );
		    name = apr_pstrndup(pool, temp, strlen(temp));
		    
		    if ( ns_id && name )
            {
                apr_hash_set ( db_r->ns_hash, name, APR_HASH_KEY_STRING, ns_id );
                DBG2 ( "ADD Hash : %s: %s", ns_id, name );
            }
		    
    	}

	}
	/* EXPERIMENTAL END */

    return;
}


/* Insert new namespace
 * Must build hash and try to find namespace first.
 */
static long dbms_insert_ns ( dav_repos_db * d,
                             dav_repos_resource * db_r, const char *namespace )
{
    long id;
    apr_pool_t *pool = db_r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0, nrows = 0, first = 1;
	/* EXPERIMENTAL END */
    
    TRACE();

    dbd = dbms_get_db(d);
    
    if (dbd) 
    {
    	/* We need to add name space to DB and fill hash */
		rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, 
				"INSERT INTO dasl_namespace (name) VALUES (%s)", 
				NULL, &pquery);
			
		if(rv != 0) {
			db_error_message_new(dbd, rv, "prepare statement failed", __func__);
			return 0;
		}
				
		rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
				namespace);
				
		if(rv != 0) {
			db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
			return 0;
		}
		
		/* Success to insert */
		/* Get ns_id from DB */
		rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, 
							 "SELECT * FROM dasl_namespace WHERE name=%s", 
		                     NULL, 
		                     &pselect);
			
		if(rv != 0) {
			db_error_message_new(dbd, rv, "prepare statement failed", __func__);
			return -1;
		}
				
		rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1,
				namespace );
		
		if(rv != 0) {
			db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
			return -1;
		}
		
		/* Get ns_id */
		for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
			
			if (first) {
				/* Add to hash_table */
			    if ( db_r->ns_hash )
			        apr_hash_set ( db_r->ns_hash, apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row, 1) ),
			                       APR_HASH_KEY_STRING, apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row, 0) ) );

			    id = atol ( apr_dbd_get_entry(dbd->driver, row, 0) );
			}
			first = 0;
		}

		return id;
	} 
    else 
    {
		return 0;
	}
    
}

/*## FIXME : Make it building hash only once per request
 *
 */
/* Get namespace
 * If new namespace, save it and return ns_id
 *
 * @param d DB connection struct
 * @param r Pointer to a resource struct identifying the resource in question
 * @param namespace namespace
 * @return name space. Should not return NULL
 * */
const char *dbms_get_ns ( dav_repos_db * d,
                          dav_repos_resource * db_r, const long ns_id )
{
    const char *namespace;
    const char *ns_id_str;
    const char *ns_default = "http://catacomb.tigris.org";

    TRACE();

    /* Build hash first */
    if ( db_r->ns_id_hash == NULL )
        return ns_default;

    /* Build key */
    ns_id_str = apr_psprintf ( db_r->p, "%ld", ns_id );
    namespace =
                    ( const char * ) apr_hash_get ( db_r->ns_id_hash, ns_id_str,
                                                    APR_HASH_KEY_STRING );

    /* We can't return NULL, so let's return default NS */
    if ( namespace == NULL )
{
        return ns_default;
    }

    return namespace;
}

/* Get namespace
 * If new namespace, save it and return ns_id
 *
 * @param d DB connection struct
 * @param r Pointer to a resource struct identifying the resource in question
 * @param namespace namespace
 * @return ns_id
 * */
long dbms_get_ns_id ( dav_repos_db * d,
                      dav_repos_resource * db_r, const char *namespace )
{


    char *ns_id = NULL;

    TRACE();

    dbms_build_ns_hash ( d, db_r );

    /* Now, We have Hash table */
    ns_id = apr_hash_get ( db_r->ns_hash, namespace, APR_HASH_KEY_STRING );

    /* name space is not found */
    if ( ns_id )
        return atol ( ns_id );

    else
        return dbms_insert_ns ( d, db_r, namespace );
}

/* This function for walker */
dav_error *dbms_load_locknull_list ( dav_repos_db * db,
                                     char *path,
                                     apr_text_header * phdr, apr_pool_t * pool )
{

    /* EXPERIMENTAL START */
    
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	/* EXPERIMENTAL END */
    
    TRACE();
    dav_repos_no_trail ( path );
    DBG0 ( path );

    dbd = dbms_get_db(db);
    
    if (dbd) 
    {
		
    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT fname FROM dasl_locknull WHERE path=%s", NULL, &pselect);
    		
		if(rv != 0) {
		   	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
		   	return dav_new_error ( pool, 
		   			HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
				
		rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1, 
				path);
				
		if(rv != 0) {
		  	db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
		  	return dav_new_error ( pool, 
		  			HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
		
		for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) 
		{
			
			
			/* Add it to the text */
			apr_text_append ( pool, phdr, apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row, 0) ) );
		}
		
		return NULL;
		
	} 
    else 
    {
    	return dav_new_error ( pool, HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error - no DB handle" );
	}
}

dav_error *dbms_save_locknull_member ( dav_repos_db * db,
                                       const char *filename, apr_pool_t * pool )
{
    const char *fname = filename;
    char *path = ap_make_dirstr_parent ( pool, filename );

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	int rv = 0, nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    /* no trailing */
    dav_repos_no_trail ( path );

    dbd = dbms_get_db(db);
    
    if (dbd) 
    {
    	/* We need to add name space to DB and fill hash */
		rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "insert into dasl_locknull VALUES (%s, %s)", 
		                     NULL, 
		                     &pquery);
			
		if(rv != 0) {
		   	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
		   	return dav_new_error ( pool,
		   			HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
				
		rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
				path, 
				fname);
				
		if(rv != 0) {
			db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
		  	return dav_new_error ( pool, 
		  			HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
		
		return NULL;
		
	} 
	return dav_new_error ( pool, 
			HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
}

dav_error *dbms_remove_locknull_member ( dav_repos_db * db,
        const char *filename,
        apr_pool_t * pool )
{
    const char *fname = filename;
    char *path = ap_make_dirstr_parent ( pool, filename );

	/* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	int rv = 0, nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    /* no trailing */
    dav_repos_no_trail ( path );

    dbd = dbms_get_db(db);
    
    if (dbd)
    {
    
	    /* We need to add name space to DB and fill hash */
	    rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "DELETE FROM dasl_locknull WHERE path=%s AND fname=%s", 
		                     NULL, 
		                     &pquery);
			
		if(rv != 0) {
		    db_error_message_new(dbd, rv, "prepare statement failed", __func__);
		   	return dav_new_error ( pool, 
		   			HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
				
		rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
				path,
				fname );
				
		if(rv != 0) {
		    db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
		  	return dav_new_error ( pool, 
		  			HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}

    	return NULL;
    } 
    else 
    {
    	return dav_new_error ( pool,
    			HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
    }
}


dav_error *dbms_load_lock_record ( dav_repos_db * db,
                                   char *key,
                                   dav_lock_discovery ** direct,
                                   dav_lock_indirect ** indirect,
                                   int *need_save,
                                   apr_pool_t * pool )
{
    dav_lock_discovery *dp = NULL;
    dav_lock_indirect *ip = NULL;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_results_t * res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	/* EXPERIMENTAL END */
    

    TRACE();
    dav_repos_no_trail ( key );
    
    dbd = dbms_get_db(db);
    
    if (dbd) 
    {
		rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT locktype, scope, depth, "
                       "timeout, locktoken, owner, auth_user, lockkey "
                       "FROM dasl_lock WHERE uri=%s", NULL, &pselect);
		
		if (rv != 0) {
			db_error_message_new(dbd, rv, "error while preparing statement", __func__);
			return dav_new_error ( pool, 
								   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
		
		rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1,
				key);
		
		if (rv != 0) {
			db_error_message_new(dbd, rv, "error while execute statement", __func__);
			return dav_new_error ( pool, 
								   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
		
		/* Get Results */
		for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) 
		{
			
			dav_locktoken *locktoken;

	        /* Check timeout */
	        if ( dav_repos_lock_expired ( atoll ( apr_dbd_get_entry(dbd->driver, row , 3) ) ) )
	        {
	            *need_save = DAV_TRUE;
	            continue;
	        }

	        locktoken = apr_pcalloc ( pool, sizeof ( *locktoken ) );
	        locktoken->char_uuid = apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row , 4) );

	        if ( atoi ( apr_dbd_get_entry(dbd->driver, row , 0) ) == DAV_LOCK_DIRECT )
	        {

	            dp = apr_pcalloc ( pool, sizeof ( *dp ) );

	            /* Fill DP */
	            dp->f.scope = atoi ( apr_dbd_get_entry(dbd->driver, row , 1) );
	            dp->f.depth = atoi ( apr_dbd_get_entry(dbd->driver, row , 2) );
	            dp->f.timeout = atoll ( apr_dbd_get_entry(dbd->driver, row , 3) );

	            dp->locktoken = locktoken;

	            dp->owner = apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row , 5) );
	            if ( apr_dbd_get_entry(dbd->driver, row , 7) && strlen ( apr_dbd_get_entry(dbd->driver, row , 6) ) )
	                dp->auth_user = apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row , 6) );

	            dp->next = *direct;
	            *direct = dp;
	        }
	        else
	        {
	            ip = apr_pcalloc ( pool, sizeof ( *ip ) );

	            /* Fill IP */
	            ip->timeout = atoll ( apr_dbd_get_entry(dbd->driver, row , 3) );
	            ip->locktoken = locktoken;
	            ip->key = apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row , 7) );

	            /* Put it in the chain */
	            ip->next = *indirect;
	            *indirect = ip;
	        }
	    }
		
		return NULL;
	} 
    else 
    {
    	return dav_new_error ( pool, 
    						   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
	}
}


/* DIRECT LOCK save
** lock info has linked list data
** #FIXME : Need rollback if we had error
*/
dav_error *dbms_save_lock_record ( dav_repos_db * db,
                                   char *key,
                                   const dav_lock_discovery * direct,
                                   const dav_lock_indirect * indirect,
                                   apr_pool_t * pool )
{
    dav_error *err;
    const dav_lock_discovery *dp;
    const dav_lock_indirect *ip;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	apr_dbd_transaction_t *trans = NULL;
	int rv = 0, nrows = 0, dberr = 0;
	/* EXPERIMENTAL END */
    
    TRACE();
    dav_repos_no_trail ( key );

    if ( direct == NULL && indirect == NULL )
    {
        DBG0 ( "Nothing to save" );
        return NULL;
    }

    /* FIXME: Should I remove all data??
    ** Should we find better way?
    */
    err = dbms_remove_lock_record ( db, key ,pool );
    if ( err != NULL )
    {
        return err;
    }

    dbd = dbms_get_db(db);
    
    if (dbd)
    {
    	/* start transaction */
    	rv = apr_dbd_transaction_start(dbd->driver, pool, dbd->handle, &trans);
    	
    	if (rv != 0) {
			db_error_message_new(dbd, rv, "Transaction start failed", __func__);
		}
    	
		for ( dp = direct; dp && !dberr; dp = dp->next )
		{
			
			rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "INSERT INTO dasl_lock "
					"(URI, locktype, scope, depth, timeout, locktoken, owner, auth_user) "
					"VALUES (%s, %d, %d, %d, %d, %s, %s, %s)", 
			        NULL, 
			        &pquery);
				
			if(rv != 0) {   	
			   	db_error_message_new(dbd, rv, "prepare insert statement failed", __func__);
				dberr = -1;
			}
			
			rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
				    key, 
				    apr_itoa( pool, DAV_LOCK_DIRECT ), 
				    apr_itoa( pool, dp->f.scope ), 
				    apr_itoa( pool, dp->f.depth ), 
				    apr_ltoa( pool, dp->f.timeout ), 
				    dav_repos_dbms_string( dp->locktoken->char_uuid ), 
				    dav_repos_dbms_string( dp->owner ), 
				    dav_repos_dbms_string( dp->auth_user ));
			
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepared insert statement execution failed", __func__);
				dberr = -1;
			}
			
		}
		
		
		for ( ip = indirect; ip && !dberr; ip = ip->next )
		{
			
			rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, 
					"insert into dasl_lock \
		            (URI, locktype, timeout, locktoken, lockkey) \
		            VALUES (%s, %d, %d, %s, %s)", 
                    NULL, 
                    &pquery);
				
			if(rv != 0) {   	
			   	db_error_message_new(dbd, rv, "prepare insert statement failed", __func__);
				dberr = -1;
			}
					
			rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
					key, 
				    apr_itoa( pool, DAV_LOCK_INDIRECT ), 
				    apr_ltoa( pool, ip->timeout ), 
				    dav_repos_dbms_string( ip->locktoken->char_uuid ), 
				    dav_repos_dbms_string( ip->key ));
					
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepared insert statement execution failed", __func__);
				dberr = -1;
			}
			
		}
		
		rv = apr_dbd_transaction_end(dbd->driver, pool, trans);
		
		if (rv != 0)
		{
			db_error_message_new(dbd, rv, "Error occurred in the transaction", __func__);
			return dav_new_error ( pool, 
								   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
		
		return NULL;
    }
    else
    {
    	return dav_new_error ( pool, 
    						   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
    }
}

dav_error *dbms_remove_lock_record ( dav_repos_db * db,
                                     char *key, apr_pool_t * pool )
{

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	int rv = 0, nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();
    dav_repos_no_trail ( key );

    dbd = dbms_get_db(db);
    
    if (dbd)
    {
    	/* We need to add name space to DB and fill hash */
    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, 
    						 "DELETE FROM dasl_lock WHERE uri=%s", 
		                     NULL, 
		                     &pquery);
			
		if(rv != 0) {
		    db_error_message_new(dbd, rv, "prepare statement failed", __func__);
		   	return dav_new_error ( pool, 
		   			HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
				
		rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
				key );
				
		if(rv != 0) {
		    db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
		  	return dav_new_error ( pool, 
		  			HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
		
    	return NULL;
    }
    else
    {
    	return dav_new_error ( pool, 
    			HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
    }
    
}


/*
** FIXME : Can we just count without get the result??
*/
int dbms_lock_exists ( dav_repos_db * db,
                       char *key, apr_pool_t * pool )
{

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0, nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();
    dav_repos_no_trail ( key );

    dbd = dbms_get_db(db);
    
    if (dbd) 
    {
    	/*Create the search command */
    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, 
    			"SELECT COUNT(uri) FROM dasl_lock WHERE uri=%s", 
    			NULL, &pselect);
    	
    	if (rv != 0) {
			db_error_message_new(dbd, rv, "preparing Select Statement", __func__);
			return -1;
		}
    	
    	rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1, 
    			key);
    	
    	if (rv != 0) {
			db_error_message_new(dbd, rv, "execute Select Statement", __func__);
			return -1;
		}
    	
    	for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
			nrows = atoi(apr_dbd_get_entry(dbd->driver, row, 0));
		}
    	
    	if ( nrows == 0 )
    		return 0;
    	else if (nrows > 0)
            return 1;
    	else 
    		return -1;

	} 
    else 
    {
		return -1;
	}
    
}               /*End of dbms_exists */

/**
 * Trun on the versioning for a resource
 * @param d DB connection struct containing the user, password, and DB name
 * @param db_r Identifies the resource to set the property on r->uri.
 *          Contains the live property and value to set.
 * @checkin: Checkin version
 * @return NULL, for success; Error, otherwise.
 */
dav_error * dbms_set_is_versioned ( dav_repos_db * db, dav_repos_resource * db_r,
                                    int checkin )
{
    apr_pool_t *pool = db_r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	int rv = 0, nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    dbd = dbms_get_db(db);
    
    if (dbd) 
    {
		rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "UPDATE dasl_resource SET isversioned=1,"
                       		 "checkin=%d WHERE serialno=%d", 
		                     NULL, 
		                     &pquery);
			
		if(rv != 0) {   	
		   	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
		   	return dav_new_error ( pool, 
		   						   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
				
		rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
				apr_itoa( pool, checkin ),
				apr_ltoa( pool, db_r->serialno ) );
				
		if(rv != 0) {
			db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
			return dav_new_error ( pool,
								   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}

    	return NULL;
	} 
    else 
    {
    	return dav_new_error ( pool,
    						   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
	}
     
}               /*End of dbms_set_is_versioned*/

/**
 * Set the checkin and checkout version for a resource
 * @param d DB connection struct containing the user, password, and DB name
 * @param db_r Identifies the resource to set the property on r->uri.
 *          Contains the live property and value to set.
 * @checkin: Checkin version
 * @checkout: Checkout version
 * @return NULL, for success; Error, otherwise.
 */
dav_error * dbms_set_checkin_out ( dav_repos_db * db, dav_repos_resource * db_r,
                                   int checkin, int checkout )
{
    apr_pool_t *pool = db_r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	int rv = 0, nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    dbd = dbms_get_db(db);
    
    if (dbd) {
		
    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "UPDATE dasl_resource SET checkin=%d,"
                       		 "checkout=%d WHERE serialno=%d", 
		                     NULL, 
		                     &pquery);
			
		if(rv != 0) {   	
		   	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
		   	return dav_new_error ( pool, 
		   						   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
				
		rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
				apr_itoa( pool, checkin ),
				apr_itoa( pool, checkout ),
				apr_ltoa( pool, db_r->serialno ) );
				
		if(rv != 0) {
			db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
			return dav_new_error ( pool,
								   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
		
    	return NULL;
	} 
    else 
    {
    	return dav_new_error ( pool, 
    						   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
	}

}

/**
 * Set the checkin and checkout version for a resource
 * @param d DB connection struct containing the user, password, and DB name
 * @param db_r Identifies the resource to set the property on r->uri.
 *          Contains the live property and value to set.
 * @author: The author who changed the resource
 * @return NULL, for success; Error, otherwise.
 */
dav_error * dbms_create_vr ( dav_repos_db * db, dav_repos_resource * db_r,
                             int version )
{
    apr_pool_t *pool = db_r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pselect = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0, nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    dbd = dbms_get_db(db);
    
    if (dbd) 
    {
    	/* Drop the temporary table first */
    	rv = apr_dbd_query(dbd->driver, dbd->handle, &nrows, "DROP TABLE tmp_resforvcr");
		
    	/*
		if(rv != 0) {
			db_error_message_new(dbd, rv, "statement execution failed");
		}*/
    	
    	/*Copy the record in the 'dasl_resource' table for 'version_resource' table */

	    /*Store the source record into a temporary table */
    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT CREATE_TEMPORARY_VERSION_RESOURCE_TABLE(%d, %d, %d)", 
		                     NULL, 
		                     &pselect);
			
		if(rv != 0) {   	
		   	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
		   	return dav_new_error ( pool, 
		   						   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
				
		rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1,
				apr_ltoa(pool, db_r->serialno),
				apr_ltoa(pool, apr_time_now()),
				apr_itoa(pool, version) );
				
		if(rv != 0) {
			db_error_message_new(dbd, rv, "Error in the create temporary table command", __func__);
			return dav_new_error ( pool,
			                       HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
		
		for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
			
		}
		
		/* Duplicate the record in the temporary table
		to the table version_resource */
		rv = apr_dbd_query(dbd->driver, dbd->handle, &nrows, "INSERT INTO version_resource "
                       "SELECT * FROM tmp_resforvcr");
		
    	
		if(rv != 0) {
			db_error_message_new(dbd, rv, "Error in the insert command", __func__);
			return dav_new_error ( pool, 
								   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
		
	    if ( db_r->isexternal!=0 )
	    { // create a version of the file
	        char *src_fn = NULL, *des_fn = NULL;
	        if ( generate_path ( &src_fn, pool, db->file_dir,
	                             db_r->serialno, 0 ) )
	            return dav_new_error ( pool,
	                                   HTTP_INTERNAL_SERVER_ERROR, 0, "Unable to generate src fn" );

	        if ( generate_path ( &des_fn, pool, db->file_dir,
	                             db_r->serialno, version ) )
	            return dav_new_error ( pool,
	                                   HTTP_INTERNAL_SERVER_ERROR, 0, "Unable to generate src fn" );

	        if ( version != 0 )
	        {
	            if ( apr_file_copy ( src_fn, des_fn, APR_OS_DEFAULT, db_r->p )
	                    != APR_SUCCESS )
	                return dav_new_error ( pool,
	                                       HTTP_INTERNAL_SERVER_ERROR, 0, "Unable to copy file." );
	        }
	    }
	    
	    /*Copy the record in the 'dasl_property' table for 'version_property' table */

        /* Drop the temporary table first */
    	rv = apr_dbd_query(dbd->driver, dbd->handle, &nrows, "DROP TABLE tmp_propertyforvcr");
		
    	/*
		if(rv != 0) {
			db_error_message_new(dbd, rv, "statement execution failed");
		}*/
    	
    	/*Copy the record in the 'dasl_resource' table for 'version_resource' table */

    	/*Store the property record into a temporary table */
    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT CREATE_TEMPORARY_VERSION_PROPERTY_TABLE(%d, %d)", 
		                     NULL, 
		                     &pselect);
			
		if(rv != 0) {   	
		   	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
		   	return dav_new_error ( pool, 
		   						   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
				
		res = NULL;
		rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1,
				apr_itoa(pool, version),
				apr_ltoa(pool, db_r->serialno) );
				
		if(rv != 0) {
			db_error_message_new(dbd, rv, "Error in the create temporary table command", __func__);
			return dav_new_error ( pool,
			                       HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
	    
		for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
			
		}
		
		/* Duplicate the records in the temporary table
	    to the table version_property */
		rv = apr_dbd_query(dbd->driver, dbd->handle, &nrows, "INSERT INTO version_property "
                "SELECT * FROM tmp_propertyforvcr");
		
    	
		if(rv != 0) {
			db_error_message_new(dbd, rv, "Error in the insert command", __func__);
			return dav_new_error ( pool, 
								   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
		
		return NULL;
		
	} 
    else 
    {
    	return dav_new_error ( pool, 
    						   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
	}

}

/**
 * Find the versions of a resource
 * @param d DB connection struct containing the user, password, and DB name
 * @param db_r Identifies the resource to set the property on r->uri.
 *          Contains the live property and value to set.
 * @vcrs: Versioned resource link
 * @return NULL, for success; Error, otherwise.
 */
dav_error * dbms_get_vrs ( dav_repos_db * db, dav_repos_resource * db_r,
                           dav_repos_resource ** vrs )
{
    struct dav_repos_resource *presult_link_tail = NULL;
    struct dav_repos_resource *pnew_link_item = NULL;
    apr_pool_t *pool = db_r->p;

    /* EXPERIMENTAL START */
	ap_dbd_t * dbd = NULL;
	apr_dbd_prepared_t * pselect = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    *vrs = NULL;

    dbd = dbms_get_db(db);
    
    if (dbd) {
    	
		/*Create the search command */
    	rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "SELECT serialno, "
                       "version, "
                       "uri, "
                       "creationdate, "
                       "displayname, "
                       "getcontentlanguage, "
                       "getcontentlength, "
                       "getcontenttype, "
                       "getetag, "
                       "getlastmodified, "
                       "resourcetype, "
                       "source, "
                       "depth, "
                       "istext "
                       "FROM version_resource WHERE serialno = %d "
                       "ORDER BY version DESC", NULL, &pselect);
    	
    	if (rv != 0) {
    		db_error_message_new(dbd, rv, "Error while preparing SQL Statement", __func__);
    		return dav_new_error ( pool,
    							   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
    	
    	rv = apr_dbd_pvselect(dbd->driver, pool, dbd->handle, &res, pselect, 1, 
    			apr_ltoa(pool, db_r->serialno));
    	
    	if (rv != 0) {
			db_error_message_new(dbd, rv, "Error while execution SQL Statement", __func__);
			return dav_new_error ( pool,
			    				   HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
		}
    	
    	for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) 
    	{
	        /*fetch every row */
	
	        /*Malloc a new structure */
	        pnew_link_item = apr_pcalloc ( db_r->p, sizeof ( *pnew_link_item ) );
	        if ( presult_link_tail == NULL )
	        { /* The first one */
	            *vrs = pnew_link_item;
	            presult_link_tail = pnew_link_item;
	            presult_link_tail->next = NULL;
	        }
	        else
	        { /* Not the first one */
	            presult_link_tail->next = pnew_link_item;
	            presult_link_tail = pnew_link_item;
	        }
	
	        /*Get the data of every column of this row to a link */
	
	        presult_link_tail->serialno = atol ( apr_dbd_get_entry(dbd->driver, row, 0) );
	
	        presult_link_tail->version = atol ( apr_dbd_get_entry(dbd->driver, row, 1) );
	
	        presult_link_tail->uri = apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row, 2) );
	        DBG1 ( "CKNIGHT DEBUG: %s", presult_link_tail->uri );
	
	        presult_link_tail->creationdate = atoll ( apr_dbd_get_entry(dbd->driver, row, 3) );
	
	        presult_link_tail->displayname = apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row, 4) );
	
	        presult_link_tail->getcontentlanguage = apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row, 5) );
	
	        presult_link_tail->getcontentlength = atol ( apr_dbd_get_entry(dbd->driver, row, 6) );
	        presult_link_tail->getcontenttype =  apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row, 7) );
	
	        presult_link_tail->getetag = apr_pstrdup ( pool, apr_dbd_get_entry(dbd->driver, row, 8) );
	
	        presult_link_tail->getlastmodified = atoll ( apr_dbd_get_entry(dbd->driver, row, 9) );
	
	        presult_link_tail->resourcetype = atoi ( apr_dbd_get_entry(dbd->driver, row, 10) );
	
	        presult_link_tail->source = atoi ( apr_dbd_get_entry(dbd->driver, row, 11) );
	
	        presult_link_tail->depth = atoi ( apr_dbd_get_entry(dbd->driver, row, 12) );
	
	        presult_link_tail->istext = atoi ( apr_dbd_get_entry(dbd->driver, row, 13) );
	
	        presult_link_tail->next = NULL;
	        presult_link_tail->pr = NULL;

    	}/* End for */
    	    	
	} 
    else 
    {
    	return dav_new_error ( pool,
    	                       HTTP_INTERNAL_SERVER_ERROR, 0, "DBMS Error" );
	}

    return NULL;
}

struct dbms_stream
{
    dav_repos_db *db;
    dav_repos_resource *db_r;
    apr_file_t *file;
    char *path;
    int temporary;
};

dav_error *dbms_open_stream ( dav_repos_db *db, dav_repos_resource *db_r,
                              dbms_stream **stream, dav_stream_mode mode )
{
    dav_error *err = NULL;
    dbms_stream *s = NULL;
    apr_pool_t *pool = db_r->p;
    
    /* EXPERIMENTAL START */
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	int rv = 0, nrows = 0;
	/* EXPERIMENTAL END */
    
    TRACE();

    s = apr_pcalloc ( db_r->p, sizeof ( *s ) );
    s->db = db;
    s->db_r = db_r;

    // only way to get a serialno is to add a row (or convert to using a
    // sequence-generator
    if ( !db_r->serialno )
    {
        long serialno;
        if ( db_insert_resource ( db, db_r, &serialno ) )
            return dav_new_error ( db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                                   "Unable to add resource to the database" );
        db_r->serialno = serialno;
    }

    DBG1 ( "Max_Filesize: %li", db->max_file_size );
    DBG1 ( "Actual_contentlength: %li", db_r->getcontentlength );

    // save only in filesystem if max_file_size is smaller than contentlength
    if ( ( db->file_dir != NULL ) && ( db->max_file_size < db_r->getcontentlength ) )
    {
        if ( ( err = generate_path ( & ( s->path ), db_r->p, db->file_dir,
                                     db_r->serialno, db_r->version ) ) )
        {
            DBG0 ( "error in generating path, removing resource record." );
            
            rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "DELETE FROM dasl_resource "
                                 "WHERE serialno=%d", 
			                     NULL, 
			                     &pquery);
				
			if(rv != 0) {   	
			   	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
				return dav_new_error ( pool, HTTP_INTERNAL_SERVER_ERROR,
	   	                0, "DBMS error" );
			}
					
			rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
					apr_ltoa(pool, db_r->serialno) );
					
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
				return dav_new_error ( pool, HTTP_INTERNAL_SERVER_ERROR,
	   	                0, "DBMS error" );
			}

        }
    }

    if ( !s->path )
    {
        s->temporary = 1;
        s->path = apr_psprintf ( db_r->p, "%s/catacomb-%ld-%d",
                                 db->tmp_dir, db_r->serialno, db_r->version );
        ap_no2slash ( s->path );
    }

    DBG1 ( "path: %s", s->path );

    switch ( mode )
    {
        case DAV_MODE_WRITE_TRUNC:
        case DAV_MODE_WRITE_SEEKABLE:
            if ( apr_file_open ( & ( s->file ), s->path,
                                 APR_CREATE | APR_WRITE | APR_TRUNCATE | APR_BINARY | APR_BUFFERED,
                                 APR_OS_DEFAULT, db_r->p ) != APR_SUCCESS )
                err = dav_new_error ( db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                                      "Unable to open file for write" );
            break;
        default: // any possibility of this happening?
            if ( apr_file_open ( & ( s->file ), s->path,
                                 APR_READ | APR_BINARY | APR_BUFFERED,
                                 APR_OS_DEFAULT, db_r->p ) != APR_SUCCESS )
                err = dav_new_error ( db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                                      "Unable to open file for read" );
            break;
    }

    if ( err )
    {
        DBG0 ( "error in opening file" );
        rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "DELETE FROM dasl_resource "
                             "WHERE serialno=%d", 
		                     NULL, 
		                     &pquery);
			
		if(rv != 0) {   	
		   	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
			return dav_new_error ( pool, HTTP_INTERNAL_SERVER_ERROR,
   	                0, "DBMS error" );
		}
				
		rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
				apr_ltoa(pool, db_r->serialno) );
				
		if(rv != 0) {
			db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
			return dav_new_error ( pool, HTTP_INTERNAL_SERVER_ERROR,
   	                0, "DBMS error" );
		}
        return err;
    }

    *stream = s;
    return NULL;
}

dav_error *dbms_close_stream ( dbms_stream *stream, int commit )
{
    // TODO: merge dbms_write_content into this
    dav_error *err = NULL;

    TRACE();
    
    /* EXPERIMENTAL START */
    apr_pool_t *pool = stream->db_r->p;
	ap_dbd_t *dbd = NULL;
	apr_dbd_prepared_t *pquery = NULL;
	int rv = 0, nrows = 0;
	/* EXPERIMENTAL END */

	dbd = dbms_get_db(stream->db);
	
	if (dbd)
	{
		if ( commit )
		{
		    if ( apr_file_close ( stream->file ) == APR_SUCCESS )
		    {
		        if ( stream->temporary )
		        {
		            if ( dbms_write_content ( stream->db, stream->db_r,
		                                      stream->path ) )
		            {
		                err = dav_new_error ( stream->db_r->p,
		                                      HTTP_INTERNAL_SERVER_ERROR, 0, "Unable to close file!" );
		            }
		        }
		    }
		    else err = dav_new_error ( stream->db_r->p,
		                                   HTTP_INTERNAL_SERVER_ERROR, 0, "Unable to close file!" );
		}
		
		if ( !commit )
		{
			
			rv = apr_dbd_prepare(dbd->driver, pool, dbd->handle, "DELETE FROM dasl_resource "
		                       "WHERE serialno=%d", 
			                     NULL, 
			                     &pquery);
				
			if(rv != 0) {   	
			   	db_error_message_new(dbd, rv, "prepare statement failed", __func__);
			   	err = dav_new_error ( pool, HTTP_INTERNAL_SERVER_ERROR,
			   	                0, "DBMS error" );
			}
					
			rv = apr_dbd_pvquery(dbd->driver, pool, dbd->handle, &nrows, pquery, 
					apr_ltoa(pool, stream->db_r->serialno) );
					
			if(rv != 0) {
				db_error_message_new(dbd, rv, "prepared statement execution failed", __func__);
				err = dav_new_error ( pool, HTTP_INTERNAL_SERVER_ERROR,
							   	                0, "DBMS error" );
			}
			
		}
		
		if ( stream->temporary )
		{
		    if ( apr_file_remove ( stream->path, stream->db_r->p ) != APR_SUCCESS )
		        err = dav_new_error ( stream->db_r->p, HTTP_INTERNAL_SERVER_ERROR,
		                              0, "Unable to remove temporary file." );
		}
		return err;
	}
	else
	{
		return dav_new_error ( pool, HTTP_INTERNAL_SERVER_ERROR,
                0, "Unable to connect to database." );
	}
}

dav_error *dbms_write_stream ( dbms_stream *stream, const void *buf,
                               const apr_size_t bufsize )
{
    apr_size_t s = bufsize;
    TRACE();

    DBG1 ( "stream = %p", stream );
    DBG1 ( "stream->file = %p", stream->file );
    DBG1 ( "contentlength_of_file = %li", stream->db_r->getcontentlength );
    if ( apr_file_write ( stream->file, buf, &s ) != APR_SUCCESS )
        return dav_new_error ( stream->db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "Unable to write to file." );
    if ( s != bufsize )
        return dav_new_error ( stream->db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "Did not write all contents." );
    return NULL;
}

dav_error *dbms_seek_stream ( dbms_stream *stream, const apr_size_t pos )
{
    apr_size_t p = pos;
    TRACE();

    if ( apr_file_write ( stream->file, APR_SET, &p ) != APR_SUCCESS )
        return dav_new_error ( stream->db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "Unable to seek in file." );
    if ( p != pos )
        return dav_new_error ( stream->db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                               "Seek resulted in different position." );
    return NULL;
}

dav_error *dbms_deliver ( dav_repos_db *db, dav_repos_resource *db_r,
                          ap_filter_t *output )
{
    dav_error *err = NULL;
    char *filename = NULL;
    apr_file_t *fd;
    apr_bucket_brigade *bb;
    apr_bucket *bkt;

    if ( ( db->file_dir ) && ( db->max_file_size < db_r->getcontentlength ) )
    {
        err = generate_path ( &filename, db_r->p, db->file_dir,db_r->serialno, db_r->version );
        if ( err )
            return err;
    }
    else
    {
        /* Anyway we need to make random filename */
        filename = apr_psprintf ( db_r->p, "%s/dav_repos_deliver_XXXXXX",
                                  db->tmp_dir );
        if ( mktemp ( filename ) == NULL )
        {
            return dav_new_error ( db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                                   "An error occurred while opening a resource name." );
        }

        /* get file from data base */
        if ( dbms_read_content ( db, db_r, filename ) != 0 )
        {
            return dav_new_error ( db_r->p, HTTP_INTERNAL_SERVER_ERROR, 0,
                                   "An error occurred while getting a resource name." );
        }
    }

    DBG1 ( "TESTCDK: %s", filename );
    if ( apr_file_open ( &fd, filename, APR_READ | APR_BINARY, 0,
                         db_r->p ) != APR_SUCCESS )
    {
        return dav_new_error ( db_r->p, HTTP_FORBIDDEN, 0,
                               "File permissions deny server access." );
    }

    bb = apr_brigade_create ( db_r->p, output->c->bucket_alloc );

    /* ### this does not handle large files. but this is test code anyway */
    bkt = apr_bucket_file_create ( fd, 0,
                                   ( apr_size_t ) db_r->getcontentlength,
                                   db_r->p, output->c->bucket_alloc );
    APR_BRIGADE_INSERT_TAIL ( bb, bkt );

    bkt = apr_bucket_eos_create ( output->c->bucket_alloc );
    APR_BRIGADE_INSERT_TAIL ( bb, bkt );

    if ( ap_pass_brigade ( output, bb ) != APR_SUCCESS )
        err = dav_new_error ( db_r->p, HTTP_FORBIDDEN, 0,
                              "Could not write contents to filter." );

    if ( !db->file_dir ) apr_file_remove ( filename, db_r->p ); //ignore the error?
    return err;
}

/* EXPERIMENTAL START */
long dbms_insert_id(dav_repos_db *db, const char *table, const char *column)
{
	apr_pool_t *pool = db->r->pool;
	ap_dbd_t *dbd = NULL;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int rv = 0;
	long  retv = -1;
	const char *stmt = apr_psprintf(pool, "SELECT LAST_INSERT_ID_BY_TABLENAME_AND_COLUMN('%s', '%s')", table, column);
	
	dbd = dbms_get_db(db);
		
	if(dbd)
	{
		rv = apr_dbd_select(dbd->driver, pool, dbd->handle, &res, stmt, 1);
		
		/* Loop over der ResultSet */
		for(rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1);
			rv != -1;
			rv = apr_dbd_get_row(dbd->driver, pool, res, &row, -1)) {
			
			if(rv != 0) {
			    db_error_message_new(dbd, rv, "Error while retrieving ID", __func__);
			}
			else {
				retv = atol(apr_dbd_get_entry(dbd->driver, row, 0));
			}
		}
	}
	
	return retv;
}

ap_dbd_t *dbms_get_db(dav_repos_db *db)
{
	if (!db->dbd)
		db->dbd = ap_dbd_acquire(db->r);
	
	return db->dbd;
}
/* EXPERIMENTAL END */
