/* Copyright 2007 Deutsches Zentrum fuer Luft- und Raumfahrt e.V. (DLR)
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

/*
 * mod_dav_svn.c: an Apache mod_dav dbms-module
 *
 */

#include <httpd.h>
#include <http_config.h>
#include <http_protocol.h>
#include <http_log.h>
#include <http_core.h>      /* for ap_construct_url */
#include <mod_dav.h>

#include <apr_strings.h>
#include <apr_hash.h>

#include "dav_repos.h"
#include "dbms.h"
#include "mod_dav_repos_version.h"

#define INHERIT_VALUE(parent, child, field) \
                ((child)->field ? (child)->field : (parent)->field)

#define DBMS_PRIVIDER_NAME "repos"

/* Note: the "dav_repos" prefix is mandatory */
extern module AP_MODULE_DECLARE_DATA dav_repos_module;

static void *dav_repos_create_dir_config ( apr_pool_t * p, char *dir )
{
    /* NOTE: dir==NULL creates the default per-dir config */
    dav_repos_dir_conf *conf;

    conf = ( dav_repos_dir_conf * ) apr_pcalloc ( p, sizeof ( *conf ) );

    return conf;
}

static void *dav_repos_merge_dir_config ( apr_pool_t * p, void *base,
        void *overrides )
{
    /** FIXME : for future use
    dav_repos_dir_conf *parent = base;
    dav_repos_dir_conf *child = overrides;
    */
    dav_repos_dir_conf *newconf;

    newconf = apr_pcalloc ( p, sizeof ( *newconf ) );
    return newconf;
}

static void *dav_repos_create_server_config ( apr_pool_t * p, server_rec * s )
{
    dav_repos_server_conf *newconf;
    newconf = ( dav_repos_server_conf * ) apr_pcalloc ( p, sizeof ( *newconf ) );
    return newconf;
}

static void *dav_repos_merge_server_config ( apr_pool_t * p, void *base,
        void *overrides )
{
    dav_repos_server_conf *parent = base;
    dav_repos_server_conf *child = overrides;

    dav_repos_server_conf *newconf;
    newconf = ( dav_repos_server_conf * ) apr_pcalloc ( p, sizeof ( *newconf ) );

    newconf->tmp_dir = INHERIT_VALUE ( parent, child, tmp_dir );
    newconf->file_dir = INHERIT_VALUE ( parent, child, file_dir );
    newconf->exact_search = INHERIT_VALUE ( parent, child, exact_search );
    newconf->max_file_size = INHERIT_VALUE ( parent, child, max_file_size );

    return newconf;
}


/*
 * Command handler for the DAV directive, which is TAKE1.
 */
static const char *dav_repos_tmp_dir_cmd ( cmd_parms * cmd, void *config,
        const char *arg1 )
{
    dav_repos_server_conf *conf =
        ap_get_module_config ( cmd->server->module_config,
                               &dav_repos_module );

    conf->tmp_dir = apr_pstrdup ( cmd->pool, arg1 );
    return NULL;

}

static const char *dav_repos_file_dir_cmd ( cmd_parms * cmd, void *config,
        const char *arg1 )
{
    dav_repos_server_conf *conf =
        ap_get_module_config ( cmd->server->module_config,
                               &dav_repos_module );

    conf->file_dir = apr_pstrdup ( cmd->pool, arg1 );
    return NULL;

}

// Maximun File Size for storing in DB. This is at testing-stage.
static const char *dav_repos_max_f_size_cmd ( cmd_parms * cmd, void *config,
        const char *arg1 )
{
    dav_repos_server_conf *conf =
        ap_get_module_config ( cmd->server->module_config,
                               &dav_repos_module );

    conf->max_file_size = atol ( arg1 );
    return NULL;

}

static const char *dav_repos_exact_search_cmd ( cmd_parms *cmd, void *config, const int arg )
{
    dav_repos_server_conf *conf =
        ap_get_module_config ( cmd->server->module_config,
                               &dav_repos_module );

    conf->exact_search = arg;
    return NULL;
}

static const command_rec dav_repos_cmds[] =
    {
        /* per directory/location */
        /* how can I make it mandatory */
        /* RSRC_CONF : For server */
        /* ACCESS_CONF : For dir */
        AP_INIT_TAKE1 ( "DavDBMSTmpDir", dav_repos_tmp_dir_cmd, NULL, RSRC_CONF,
                        "specify the MYSQL_TMP_DIR for a directory or location" ),

        AP_INIT_TAKE1 ( "DavDBMSFileDir", dav_repos_file_dir_cmd, NULL, RSRC_CONF,
                        "specify the directory for permanent external storage" ),

        AP_INIT_FLAG ( "DavDBMSUseExactSearch", dav_repos_exact_search_cmd, NULL, RSRC_CONF,
                       "specify whether to use exact FULLTEXT search, or the relevancy heuristics of MySQL" ),

        AP_INIT_TAKE1 ( "DavDMBSMaxFileSize", dav_repos_max_f_size_cmd, NULL, RSRC_CONF,
                        "specify the directory for permanent external storage" ),

        {NULL}
    };

/** Add DASL/0.1 version */
static int dav_repos_init_handler ( apr_pool_t * p, apr_pool_t * plog,
                                    apr_pool_t * ptemp, server_rec * s )
{
    DBG0 ( "init_handler" );

    ap_add_version_component ( p, "Catacomb/" VERSION );
    return OK;
}

/* Provider interface */
static const dav_provider dav_repos_provider =
    {
        &dav_repos_hooks_repos, /* repos */
        &dav_repos_hooks_propdb,    /* prop */
        &dav_repos_hooks_locks, /* locks */
        &dav_repos_hooks_vsn,       /* versioning */
        NULL,           /* binding */
        &dav_repos_hooks_search /* search */
    };

/*
 * Closing database when httpd exit
 * TODO: is the DB var really necessary
 */
static apr_status_t dav_repos_child_exit ( void *data )
{
    server_rec *s = data;
    dav_repos_db *db =
        ap_get_module_config ( s->module_config, &dav_repos_module );

    /** FIXME: Error handling?? */
    //dbms_closedb ( db );

    /** FIXME : what should I return? */
    return 0;
}

/*
 * Open database when httpd start
 * TODO: is the DB var really necessary
 */
static void dav_repos_child_init ( apr_pool_t * p, server_rec * s )
{
    dav_repos_db *db =
        ap_get_module_config ( s->module_config, &dav_repos_module );

    /*#FIXME
       ** open db
       ** How can I report the Error??
     */
    //dbms_opendb ( db );

    /* register child_exit */
    apr_pool_cleanup_register ( p, s, dav_repos_child_exit,
                                dav_repos_child_exit );
}


static void register_hooks ( apr_pool_t * p )
{
    /* apache hooks */
    ap_hook_post_config ( dav_repos_init_handler, NULL, NULL, APR_HOOK_MIDDLE );
    ap_hook_child_init ( dav_repos_child_init, NULL, NULL, APR_HOOK_MIDDLE );

    /* live property handling */
    dav_hook_find_liveprop ( dav_repos_find_liveprop, NULL, NULL,
                             APR_HOOK_MIDDLE );
    dav_hook_insert_all_liveprops ( dav_repos_insert_all_liveprops, NULL, NULL,
                                    APR_HOOK_MIDDLE );

    /* register the namespace URIs */
    dav_register_liveprop_group ( p, &dav_repos_liveprop_group );

    /* register our provider */
    dav_register_provider ( p, DBMS_PRIVIDER_NAME, &dav_repos_provider );
}

module DAV_DECLARE_DATA dav_repos_module = {
            STANDARD20_MODULE_STUFF,
            dav_repos_create_dir_config,    /* dir config creater */
            dav_repos_merge_dir_config, /* dir merger --- default is to override */
            dav_repos_create_server_config, /* server config */
            dav_repos_merge_server_config,  /* merge server config */
            dav_repos_cmds,     /* command table */
            register_hooks,     /* register hooks */
        };
