/* Copyright 2007 Deutsches Zentrum fuer Luft- und Raumfahrt e.V. (DLR)
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#include <httpd.h>
#include <http_config.h>
#include <http_protocol.h>
#include <http_log.h>
#include <http_request.h>
#include <http_core.h>      /* for ap_construct_url */
#include <mod_dav.h>

#include <apr.h>
#include <apr_strings.h>
#include <apr_hash.h>
#include <apr_tables.h>
#include <apr_file_io.h>

#include "dav_repos.h"
#include "dbms.h"
#include "util.h"
#include "version.h"

/* EXPERIMENTAL START */
#include <apr_dbd.h>
#include <mod_dbd.h>
/* EXPERIMENTAL END */

/* Note: the "dav_repos" prefix is mandatory */
extern module AP_MODULE_DECLARE_DATA dav_repos_module;

dav_repos_db *dav_repos_get_db ( request_rec * r )
{
    if ( r == NULL )
        return NULL;

    return ( dav_repos_db * )
           ap_get_module_config ( r->server->module_config, &dav_repos_module );
}


/* Note: picked up from ap_gm_timestr_822() */
/* NOTE: buf must be at least DAV_TIMEBUF_SIZE chars in size */
void dav_repos_format_time ( int style, apr_time_t sec, char *buf )
{
    apr_time_exp_t tms;

    /* ### what to do if fails? */
    ( void ) apr_time_exp_gmt ( &tms, sec );

    if ( style == DAV_STYLE_ISO8601 )
    {
        /* ### should we use "-00:00" instead of "Z" ?? */

        /* 20 chars plus null term */
        sprintf ( buf, "%.4d-%.2d-%.2dT%.2d:%.2d:%.2dZ",
                  tms.tm_year + 1900, tms.tm_mon + 1, tms.tm_mday,
                  tms.tm_hour, tms.tm_min, tms.tm_sec );
        return;
    }

    /* RFC 822 date format; as strftime '%a, %d %b %Y %T GMT' */

    /* 29 chars plus null term */
    sprintf ( buf,
              "%s, %.2d %s %d %.2d:%.2d:%.2d GMT",
              apr_day_snames[tms.tm_wday],
              tms.tm_mday, apr_month_snames[tms.tm_mon],
              tms.tm_year + 1900, tms.tm_hour, tms.tm_min, tms.tm_sec );
}


const char *dav_repos_build_ns_name_key ( const char *ns, const char *name,
        apr_pool_t * pool )
{
    /* Woops */
    if ( ns == NULL || name == NULL || pool == NULL )
        return NULL;

    if ( strlen ( ns ) ==0 )
        ns=" "; /* No namespace */

    return apr_psprintf ( pool, "%s\t%s", ns, name );
}

void dav_repos_get_ns_name_key ( const char *key, dav_prop_name * pname,
                                 apr_pool_t * pool )
{
    char *str;
    pname->ns = pname->name = NULL;

    TRACE();

    /* Woops */
    if ( key == NULL || pool == NULL )
        return;

    str = apr_pstrdup ( pool, key );
    pname->ns = strtok ( str, "\t" );
    pname->name = strtok ( NULL, "\t" );
}

static const char *dav_repos_find_namespace ( int ns,
        apr_array_header_t * namespaces )
{
    const char **puri = ( const char ** ) namespaces->elts;

    /* Boundary */
    if ( ns < 0 || ns > namespaces->nelts )
        return "";

    return puri[ns];
}

static const char *dav_repos_find_props ( dav_repos_resource *db_r,
        apr_xml_doc *doc,
        apr_xml_elem *elem )
{
    apr_pool_t *pool = db_r->p;

    TRACE();

    DBG2 ( "[hunkim] find <%s:%s>\n",
           dav_repos_find_namespace ( elem->ns, doc->namespaces ), elem->name );

    /* "DAV:" namespace */
    if ( elem->ns == APR_XML_NS_DAV_ID )
    {
        if ( !strcmp ( elem->name, "lockdiscovery" ) )
            return  db_r->lockdiscovery;


        if ( !strcmp ( elem->name, "supportedlock" ) )
            return db_r->supportedlock;

        if ( !strcmp ( elem->name, "displayname" ) )
            return db_r->displayname;

        return  apr_hash_get ( db_r->lpr_hash, elem->name, APR_HASH_KEY_STRING );

    }
    /* Other namespace */
    else
    {
        const char *ns = dav_repos_find_namespace ( elem->ns, doc->namespaces );
        /* Get the key */
        const char *key =
            dav_repos_build_ns_name_key ( ns, elem->name, pool );

        /* Get the dead prop */
        dav_repos_property *dead_prop =
            apr_hash_get ( db_r->pr_hash, key, APR_HASH_KEY_STRING );
        if ( dead_prop )
            return dead_prop->value;

    }

    return NULL;
}

/* Make XML response */
dav_response *dav_repos_mkresponse ( dav_repos_db * db,
                                     dav_repos_resource * db_r,
                                     apr_xml_doc *doc,
                                     apr_xml_elem *elem )
{
    int i;
    int not_found = 0;   /* flag for not found props */
    char *s = NULL;
    const char *ns;
    apr_text_header nohdr = { 0 };
    apr_text_header hdr = { 0 };
    apr_text_header hdr_ns = { 0 };
    dav_response *res = apr_pcalloc ( db_r->p, sizeof ( *res ) );
    dav_repos_property *dead_prop;
    apr_pool_t *pool = db_r->p;
    int ns_use[DAV_REPOS_MAX_NAMESPACE];    /* Maximul namespaces */

    /* Resopnse Example
       <?xml version="1.0" encoding="utf-8"?>
       <D:multistatus xmlns:D="DAV:">
       <D:response xmlns:lp1="DAV:" xmlns:lp2="http://apache.org/dav/props/">
       <D:href>/</D:href>
       <D:propstat>
       <D:prop>
       <lp1:resourcetype><D:collection/></lp1:resourcetype>
       <lp1:creationdate>2001-10-22T02:46:33Z</lp1:creationdate>
       <lp1:getlastmodified>Mon, 22 Oct 2001 02:46:33 GMT</lp1:getlastmodified>
       <lp1:getetag>"59272-1000-7e1c7440"</lp1:getetag>
       <D:supportedlock>
       <D:lockentry>
       <D:lockscope><D:exclusive/></D:lockscope>
       <D:locktype><D:write/></D:locktype>
       </D:lockentry>
       <D:lockentry>
       <D:lockscope><D:shared/></D:lockscope>
       <D:locktype><D:write/></D:locktype>
       </D:lockentry>
       </D:supportedlock>
       <D:lockdiscovery/>
       <D:getcontenttype>httpd/unix-directory</D:getcontenttype>
       </D:prop>
       <D:status>HTTP/1.1 200 OK</D:status>
       </D:propstat>
       </D:response>
       </D:multistatus>
     */

    /* Make fake data for test */
    res->href = db_r->uri;
    res->status = 200;

    TRACE();

    /* Generate namespace first */
    /* Dead properties */
    memset ( ns_use, 0x00, sizeof ( ns_use ) );
    for ( dead_prop = db_r->pr; dead_prop; dead_prop = dead_prop->next )
    {
        ns_use[ ( int ) dead_prop->ns_id] = 1;
    }

    /* Add namespace for dead properties */
    for ( i = 0; i < DAV_REPOS_MAX_NAMESPACE; i++ )
    {
        if ( ns_use[i] == 1 )
        {
            ns = dbms_get_ns ( db, db_r, i );
            if ( ns == NULL || strlen ( ns ) == 0 )
            {
                ns_use[i] = -1; /* nullns */
            }
            else
            {
                s = apr_psprintf ( pool, " xmlns:%s%d=\"%s\"", "ns", i, ns );
                apr_text_append ( pool, &hdr_ns, s );
            }
        }
    }

    /* Make XML here */

    apr_text_append ( pool, &hdr,
                      "<D:propstat>" DEBUG_CR "<D:prop>" DEBUG_CR );

    /* Start prop */
    /* All props */
    if ( elem == NULL )
    {
        DBG0 ( "[hunkim] allprops;;;;" );
        /* Dead properties */
        for ( dead_prop = db_r->pr; dead_prop; dead_prop = dead_prop->next )
        {
            /* Null ns need xmlns="" */
            if ( ns_use[ ( int ) dead_prop->ns_id] == -1 )
                s = apr_psprintf ( pool, "<%s xmlns=\"\">%s</%s>" DEBUG_CR,
                                   dead_prop->name, dead_prop->value,
                                   dead_prop->name );
            else
                s = apr_psprintf ( pool, "<ns%ld:%s>%s</ns%ld:%s>" DEBUG_CR,
                                   dead_prop->ns_id, dead_prop->name,
                                   dead_prop->value,
                                   dead_prop->ns_id, dead_prop->name );

            apr_text_append ( pool, &hdr, s );
            ns_use[ ( int ) dead_prop->ns_id] = 1;
        }

        /* Live properties */
        if ( db_r->lpr_hash )
        {   /* build it before mkresponse */
            int klen;
            const char *key, *val;
            apr_hash_index_t *hindex;

            /* Read live props from hash */
            for ( hindex = apr_hash_first ( pool, db_r->lpr_hash );
                    hindex; hindex = apr_hash_next ( hindex ) )
            {
                apr_hash_this ( hindex, ( void * ) &key, &klen, ( void * ) &val );
                s = apr_psprintf ( pool, "<D:%s>%s</D:%s>" DEBUG_CR, key, val,
                                   key );
                apr_text_append ( pool, &hdr, s );
            }
        }

    }
    else
    {
        DBG0 ( "[hunkim] props;;;;" );

        /* Print only required props */
        for ( ; elem; elem = elem->next )
        {
            const char *value = dav_repos_find_props ( db_r, doc, elem );
            if ( value == NULL )
            {
                not_found = 1;
                s = apr_psprintf ( pool, "<%s xmlns=\"%s\"/>" DEBUG_CR,
                                   elem->name,
                                   dav_repos_find_namespace ( elem->ns, doc->namespaces ) );
                apr_text_append ( pool, &nohdr, s );
            }
            else
            {
                s = apr_psprintf ( pool, "<%s xmlns=\"%s\">%s</%s>" DEBUG_CR,
                                   elem->name,
                                   dav_repos_find_namespace ( elem->ns, doc->namespaces ),
                                   value, elem->name );
                apr_text_append ( pool, &hdr, s );
            }

        }
    }

    /* Lock discovery print, if we have */
    /* FIXME : Do we really need lock info for search */
#if 0
    apr_text_append ( pool, &hdr, "<lp1:lockdiscovery>" DEBUG_CR );
    if ( db_r->lockdiscovery )
        apr_text_append ( pool, &hdr, db_r->lockdiscovery );
    apr_text_append ( pool, &hdr, "</lp1:lockdiscovery>" DEBUG_CR );

    /* Supported lock */
    if ( db_r->supportedlock )
        apr_text_append ( pool, &hdr, db_r->supportedlock );
#endif

    /* Closing 200 */
    apr_text_append ( pool, &hdr,
                      "</D:prop>" DEBUG_CR
                      "<D:status>HTTP/1.1 200 OK</D:status>" DEBUG_CR
                      "</D:propstat>" DEBUG_CR );


    /* Opening 400 */
    if ( not_found==1 )
    {
        apr_text *t;

        apr_text_append ( pool, &hdr,
                          "<D:propstat>" DEBUG_CR "<D:prop>" DEBUG_CR );


        /* Append not founded props */
        for ( t = nohdr.first; t != NULL; t = t->next )
            apr_text_append ( pool, &hdr, t->text );

        /* Closing 404 */
        apr_text_append ( pool, &hdr,
                          "</D:prop>" DEBUG_CR
                          "<D:status>HTTP/1.1 404 Not Found</D:status>" DEBUG_CR
                          "</D:propstat>" DEBUG_CR );
    }


    /*
    ** name space set up
    ** Better way??
    */
    s = apr_psprintf ( pool, " xmlns:%s%d=\"%s\"", "lp", 0, "DAV:" );
    apr_text_append ( pool, &hdr_ns, s );

    s = apr_psprintf ( pool, " xmlns:%s%d=\"%s\"", "lp", 1, "DAV:" );
    apr_text_append ( pool, &hdr_ns, s );

    res->propresult.propstats = hdr.first;
    res->propresult.xmlns = hdr_ns.first;

    return res;
}

/* Build dead property hash
 * Build the hash before makeresponse
 */
void dav_repos_build_pr_hash ( dav_repos_db * db, dav_repos_resource * db_r )
{
    const char *key;
    apr_pool_t *pool = db_r->p;
    dav_repos_property *dead_prop;

    TRACE();

    /* Let's build hash */
    db_r->pr_hash = apr_hash_make ( pool );

    /* Dead properties */
    for ( dead_prop = db_r->pr; dead_prop; dead_prop = dead_prop->next )
    {
        const char *ns = dbms_get_ns ( db, db_r, dead_prop->ns_id );
        key =
            dav_repos_build_ns_name_key ( ns, dead_prop->name, pool );
        apr_hash_set ( db_r->pr_hash, key, APR_HASH_KEY_STRING, dead_prop );
        DBG2 ( "KEY: [%s]%s", key, dead_prop->name );
    }
}

#ifdef HASH_ALL
/* Build live property */
void dav_repos_build_lpr_hash ( dav_repos_resource * db_r )
{
    const char *s;
    apr_pool_t *pool = db_r->p;

    /* an HTTP-date can be 29 chars plus a null term */
    /* a 64-bit size can be 20 chars plus a null term */
    char date[DAV_TIMEBUF_SIZE] = "";

    TRACE();

    /* Let's build hash */
    db_r->lpr_hash = apr_hash_make ( pool );


    /* #FIXME : More resourcetype for versioning */
    /* is directory */
    if ( db_r->resourcetype == dav_repos_COLLECTION )
        apr_hash_set ( db_r->lpr_hash, "resourcetype", APR_HASH_KEY_STRING,
                       "<D:collection/>" );
    else
        apr_hash_set ( db_r->lpr_hash, "resourcetype", APR_HASH_KEY_STRING, "" );

    s = apr_psprintf ( pool, "%ld", db_r->getcontentlength );
    apr_hash_set ( db_r->lpr_hash, "getcontentlength", APR_HASH_KEY_STRING, s );

    /* set live properties */
    dav_repos_format_time ( DAV_STYLE_ISO8601, db_r->creationdate, date );
    apr_hash_set ( db_r->lpr_hash, "creationdate", APR_HASH_KEY_STRING,
                   apr_pstrdup ( pool, date ) );

    dav_repos_format_time ( DAV_STYLE_RFC822, db_r->getlastmodified, date );
    apr_hash_set ( db_r->lpr_hash, "getlastmodified", APR_HASH_KEY_STRING,
                   apr_pstrdup ( pool, date ) );

    apr_hash_set ( db_r->lpr_hash, "getetag", APR_HASH_KEY_STRING,
                   dav_repos_getetag_dbr ( db_r ) );

    apr_hash_set ( db_r->lpr_hash, "getcontenttype", APR_HASH_KEY_STRING,
                   db_r->getcontenttype ? db_r->getcontenttype : "" );

    /* FIXME: Shuld it be configurable?
     * NULL for now
     */
    apr_hash_set ( db_r->lpr_hash, "getcontentlanguage",
                   APR_HASH_KEY_STRING, "" );

}

#else
/* Build live property */
void dav_repos_build_lpr_hash ( dav_repos_resource * db_r )
{
    const char *s;
    apr_pool_t *pool = db_r->p;

    /* an HTTP-date can be 29 chars plus a null term */
    /* a 64-bit size can be 20 chars plus a null term */
    char date[DAV_TIMEBUF_SIZE] = "";

    TRACE();

    /* Let's build hash */
    db_r->lpr_hash = apr_hash_make ( pool );


    /* #FIXME : More resourcetype for versioning */
    /* is directory */
    if ( db_r->resourcetype != DAV_REPOS_NODATA )
    {
        if ( db_r->resourcetype == dav_repos_COLLECTION )
            apr_hash_set ( db_r->lpr_hash, "resourcetype", APR_HASH_KEY_STRING,
                           "<D:collection/>" );
        else
            apr_hash_set ( db_r->lpr_hash, "resourcetype", APR_HASH_KEY_STRING, "" );
    }

    if ( db_r->getcontentlength != DAV_REPOS_NODATA )
    {
        s = apr_psprintf ( pool, "%ld", db_r->getcontentlength );
        apr_hash_set ( db_r->lpr_hash, "getcontentlength", APR_HASH_KEY_STRING, s );
    }

    /* set live properties */
    if ( db_r->creationdate != DAV_REPOS_NODATA )
    {
        dav_repos_format_time ( DAV_STYLE_ISO8601, db_r->creationdate, date );
        apr_hash_set ( db_r->lpr_hash, "creationdate", APR_HASH_KEY_STRING,
                       apr_pstrdup ( pool, date ) );
    }

    if ( db_r->getlastmodified != DAV_REPOS_NODATA )
    {
        dav_repos_format_time ( DAV_STYLE_RFC822, db_r->getlastmodified, date );
        apr_hash_set ( db_r->lpr_hash, "getlastmodified", APR_HASH_KEY_STRING,
                       apr_pstrdup ( pool, date ) );
    }

    /* FIXME: Always up */
    apr_hash_set ( db_r->lpr_hash, "getetag", APR_HASH_KEY_STRING,
                   dav_repos_getetag_dbr ( db_r ) );

    if ( db_r->getcontenttype )
    {
        apr_hash_set ( db_r->lpr_hash, "getcontenttype", APR_HASH_KEY_STRING,
                       db_r->getcontenttype );
    }

    /* FIXME: Shuld it be configurable?
    * NULL for now
    */
    /*
    apr_hash_set(db_r->lpr_hash, "getcontentlanguage",
    APR_HASH_KEY_STRING, "");
    */
}
#endif

/* Build version property */
/* This should not retrun for <allprop> */
void dav_repos_build_vpr_hash ( dav_repos_resource * db_r )
{
    int vnum;
    const char *vcr, *href;
    const char *vhr;
    apr_pool_t *pool = db_r->p;

    TRACE();

    /* Let's plug checkin and checkout props */
    if ( db_r->type != DAV_RESOURCE_TYPE_HISTORY &&
            db_r->type != DAV_RESOURCE_TYPE_VERSION &&
            db_r->isversioned  == 0 )
        return;

    /* Let's build version hash */
    db_r->vpr_hash = apr_hash_make ( pool );

    /* Root version */
    if ( db_r->root_version )
        href = dav_repos_mk_href ( pool, db_r->root_version );
    else
        href = dav_repos_mk_href ( pool, db_r->uri );

    apr_hash_set ( db_r->vpr_hash, "root-version",
                   APR_HASH_KEY_STRING, href );

    if ( db_r->type == DAV_RESOURCE_TYPE_REGULAR ||
            // For depth 1, the resource is type unknown
            db_r->type == DAV_RESOURCE_TYPE_UNKNOWN )
    {

        vnum = db_r->checkin==-1?
               db_r->checkout : db_r->checkin;

        /* Check in /Out status */
        if ( ( vcr = dav_repos_mk_vcr_uri ( db_r, vnum, 0 ) ) )
        {
            href = dav_repos_mk_href ( pool, vcr );
            if ( db_r->checkin != -1 )
            {
                apr_hash_set ( db_r->vpr_hash, "checked-in",
                               APR_HASH_KEY_STRING, href );
                DBG1 ( "checked-in: %s", href );
            }
            else if ( db_r->checkout != -1 )
            {
                apr_hash_set ( db_r->vpr_hash, "checked-out",
                               APR_HASH_KEY_STRING, href );
                DBG1 ( "checked-out: %s", href );
            }
        }
    }

    /* Version history resource */
    /* Make vhr by setting is_vhr 1 */
    if ( db_r->type != DAV_RESOURCE_TYPE_HISTORY )
    {
        vhr = dav_repos_mk_vcr_uri ( db_r, 0, 1 );
        if ( vhr )
        {
            href = dav_repos_mk_href ( pool, vhr );
            apr_hash_set ( db_r->vpr_hash, "version-history",
                           APR_HASH_KEY_STRING, href );
        }
    }

    /* IF is is a VR, we can compute sucessor and presendor */
    if ( db_r->type == DAV_RESOURCE_TYPE_VERSION )
    {
        int i;
        char *vname;
        char *buff;
        const char *predecessor_set;
        const char *successor_set = "";
        char date[DAV_TIMEBUF_SIZE] = "";

        buff = dav_repos_mk_vcr_uri ( db_r, db_r->version-1, 0 );
        if ( buff )
        {
            predecessor_set = dav_repos_mk_href ( pool, buff );
            /* Predecessor set */
            apr_hash_set ( db_r->vpr_hash, "predecessor-set",
                           APR_HASH_KEY_STRING, predecessor_set );
        }

        /* Add successor */
        for ( i=db_r->version+1; i<=db_r->lastversion; i++ )
        {
            buff = dav_repos_mk_vcr_uri ( db_r, i, 0 );
            if ( buff )
            {
                successor_set = apr_pstrcat ( pool, successor_set,
                                              dav_repos_mk_href ( pool,buff ),
                                              DEBUG_CR, NULL );
            }

        }

        /* Add Successor set  for hash table*/
        apr_hash_set ( db_r->vpr_hash, "successor-set",
                       APR_HASH_KEY_STRING, successor_set );

        /* Version name */
        vname = apr_psprintf ( pool, "V%d", db_r->version );
        apr_hash_set ( db_r->vpr_hash, "version-name",
                       APR_HASH_KEY_STRING, vname );


        /* Display name */
        if ( db_r->creator_displayname )
            apr_hash_set ( db_r->vpr_hash, "creator-displayname",
                           APR_HASH_KEY_STRING, db_r->creator_displayname );

        /* Comment */
        if ( db_r->comment )
            apr_hash_set ( db_r->vpr_hash, "comment",
                           APR_HASH_KEY_STRING, db_r->comment );

        /* getcontentlength */
        buff = apr_psprintf ( pool, "%ld", db_r->getcontentlength );
        apr_hash_set ( db_r->vpr_hash, "getcontentlength",
                       APR_HASH_KEY_STRING, buff );

        /* getlast modified */
        /*
        buff = apr_psprintf(pool, "%ld", db_r->getlastmodified);
        apr_hash_set(db_r->vpr_hash, "getlastmodified",
                 APR_HASH_KEY_STRING, buff);
        */

        dav_repos_format_time ( DAV_STYLE_RFC822, ( apr_time_t ) db_r->getlastmodified, date );
        apr_hash_set ( db_r->vpr_hash, "getlastmodified", APR_HASH_KEY_STRING,
                       apr_pstrdup ( pool, date ) );

    }
}

/*# FIXME : This is a insert lock info for search and prop.
** Do we really need a lock info when we search ?
** Then we should make params in search semewhere.
**
** Need to move find_live and live props
*/
dav_error *dav_repos_insert_lock_prop ( const dav_walk_params * params,
                                        dav_repos_resource * db_r )
{
    dav_error *err;
    dav_resource *resource = NULL;

    dav_walker_ctx *ctx = params->walk_ctx;

    TRACE();

    /* Initilize */
    db_r->lockdiscovery = NULL;
    db_r->supportedlock = NULL;

    if ( params->lockdb != NULL )
    {
        dav_lock *locks = NULL;

        /*
           FIXME : We don't have a resource here
           Our lock hook will use only uri and exist filed in resource
         */
        resource = apr_pcalloc ( db_r->p, sizeof ( *resource ) );
        resource->exists = 1;
        resource->uri = db_r->uri;

        if ( ( err = dav_lock_query ( params->lockdb, resource, &locks ) ) != NULL )
        {
            return dav_push_error ( db_r->p, err->status, 0,
                                    "DAV:lockdiscovery could not be "
                                    "determined due to a problem fetching "
                                    "the locks for this resource.", err );
        }

        /* fast-path the no-locks case */
        if ( locks )
        {
            /*
               ** This may modify the buffer. value may point to
               ** wb_lock.pbuf or a string constant.
             */
            db_r->lockdiscovery =
                dav_lock_get_activelock ( ctx->r, locks, NULL );

            /* Do we need strdup here */
            /* db_r->lockdiscovery = apr_pstrdup(db_r->p, wb_lock.buf); */
        }

        /*
         * Get supported lock info 
         * Our lock hook will not use resource though.
         */
        db_r->supportedlock =
            ( *params->lockdb->hooks->get_supportedlock ) ( resource );
    }

    return NULL;
}

/* make sure the pathname does not have a trailing "/" */
void dav_repos_no_trail ( char *str )
{
    int len = strlen ( str );
    while ( len > 1 && str[len - 1] == '/' )
    {
        str[len - 1] = '\0';
        len = strlen ( str );
    }
}


/* make a version resource uri using checkin number */
char *dav_repos_mk_vcr_uri ( dav_repos_resource *db_r,
                             int vnum, int is_his )
{
    char vnum_str[256];
    apr_pool_t *pool = db_r->p;
    char *parent, *filename;
    const char *uri;

    TRACE();

    /* Null URI */
    if ( db_r->uri == NULL )
        return NULL;

    /* Not a version resource */
    if ( db_r->isversioned != 1 )
        return NULL;

    /* It shouldn't happen! */
    if ( vnum == -1 )
        return NULL;

    /* If root_version, use this */
    if ( db_r->root_version )
        uri = db_r->root_version;
    else
        uri = db_r->uri;

    /* Make history version name */
    if ( is_his )
        strcpy ( vnum_str, VHR_STR );
    else
        sprintf ( vnum_str, "%d", vnum );

    parent = ap_make_dirstr_parent ( pool, uri );

    /* / case */
    if ( strlen ( uri ) == strlen ( parent ) )
        filename = apr_pstrdup ( pool, "/" );
    else
        filename = apr_pstrdup ( pool, uri + strlen ( parent ) );

    /* return new string */
    return apr_pstrcat ( pool, uri, VCR_SEPRATOR,
                         vnum_str, VCR_SEPRATOR, filename, NULL );
}

/* parse a version resource uri */
/* URI/!#!32/!#!aa.c   */
/*    ^   ^ ^          */
/* return version number
 * If it is version history is_vhr will be set
 */
char *dav_repos_parse_vcr_uri ( apr_pool_t *pool, const char *vcr_uri,
                                int *version, int *is_vhr )
{
    char *uri;
    char *end_uri, *start_version, *end_version;

    TRACE();

    if ( vcr_uri == NULL )
        return NULL;

    /* Not a VCR URI */
    if ( dav_repos_is_vcr_uri ( vcr_uri ) ==0 )
        return NULL;

    uri = apr_pstrdup ( pool, vcr_uri );

    end_uri = strstr ( uri, VCR_SEPRATOR );
    if ( end_uri==NULL )
        return NULL;

    start_version = end_uri + strlen ( VCR_SEPRATOR );
    if ( start_version==NULL )
        return NULL;

    end_version = strstr ( start_version, VCR_SEPRATOR );
    if ( end_version==NULL )
        return NULL;

    /* Make URI */
    *end_uri = 0;

    /* Make version */
    *end_version = 0;

    if ( strcmp ( start_version, VHR_STR ) )
        *is_vhr = 0;
    else
        *is_vhr = 1;

    *version = atoi ( start_version );

    return uri;
};


/* Check it is a vcr uri */
int dav_repos_is_vcr_uri ( const char *vcr_uri )
{

    /* Not a VCR URI */
    if ( strstr ( vcr_uri, VCR_SEPRATOR ) ==NULL )
        return 0;

    return 1;
}

/* Make href */
const char *dav_repos_mk_href ( apr_pool_t *pool, const char *uri )
{
    if ( uri==NULL )
        return NULL;

    return apr_pstrcat ( pool, "<D:href>",
                         uri, "</D:href>", NULL );
}

/* EXPERIMENTAL START */
char *dav_repos_lltoa(apr_pool_t *p, long long n)
{
    const int BUFFER_SIZE = sizeof(long long) * 3 + 2;
    char *buf = apr_palloc(p, BUFFER_SIZE);
    char *start = buf + BUFFER_SIZE - 1;
    int negative;
    if (n < 0) {
	negative = 1;
	n = -n;
    }
    else {
	negative = 0;
    }
    *start = 0;
    do {
	*--start = (char)('0' + (n % 10));
	n /= 10;
    } while (n);
    if (negative) {
	*--start = '-';
    }
    return start;
}

const char *dav_repos_dbms_string(const char* str)
{
	if(str != NULL) {
		return str;
	}
	return "\0";
}


/* EXPERIMENTAL END */


/**
 * Generate null resource
 */
struct dav_repos_resource *dav_repos_mk_blank_resource ( apr_pool_t *pool )
{
    dav_repos_resource *resource = apr_pcalloc ( pool, sizeof ( *resource ) );
    resource->creationdate =
    resource->getcontentlength =
    resource->serialno =
    resource->getlastmodified =
    resource->resourcetype =
    resource->source =
    resource->depth =
    resource->istext =
    resource->isexternal = DAV_REPOS_NODATA;
    return resource;
}








